using System.Collections;

using System;
using System.Text;

namespace at.jku.ssw.Analyse {



public class Parser {
	const int _EOF = 0;
	const int _ident = 1;
	const int _intCon = 2;
	const int _realCon = 3;
	const int _charCon = 4;
	const int _stringCon = 5;
	const int _abstract = 6;
	const int _as = 7;
	const int _base = 8;
	const int _bool = 9;
	const int _break = 10;
	const int _byte = 11;
	const int _case = 12;
	const int _catch = 13;
	const int _char = 14;
	const int _checked = 15;
	const int _class = 16;
	const int _const = 17;
	const int _continue = 18;
	const int _decimal = 19;
	const int _default = 20;
	const int _delegate = 21;
	const int _do = 22;
	const int _double = 23;
	const int _else = 24;
	const int _enum = 25;
	const int _event = 26;
	const int _explicit = 27;
	const int _extern = 28;
	const int _false = 29;
	const int _finally = 30;
	const int _fixed = 31;
	const int _float = 32;
	const int _for = 33;
	const int _foreach = 34;
	const int _goto = 35;
	const int _if = 36;
	const int _implicit = 37;
	const int _in = 38;
	const int _int = 39;
	const int _interface = 40;
	const int _internal = 41;
	const int _is = 42;
	const int _lock = 43;
	const int _long = 44;
	const int _namespace = 45;
	const int _new = 46;
	const int _null = 47;
	const int _object = 48;
	const int _operator = 49;
	const int _out = 50;
	const int _override = 51;
	const int _params = 52;
	const int _private = 53;
	const int _protected = 54;
	const int _public = 55;
	const int _readonly = 56;
	const int _ref = 57;
	const int _return = 58;
	const int _sbyte = 59;
	const int _sealed = 60;
	const int _short = 61;
	const int _sizeof = 62;
	const int _stackalloc = 63;
	const int _static = 64;
	const int _string = 65;
	const int _struct = 66;
	const int _switch = 67;
	const int _this = 68;
	const int _throw = 69;
	const int _true = 70;
	const int _try = 71;
	const int _typeof = 72;
	const int _uint = 73;
	const int _ulong = 74;
	const int _unchecked = 75;
	const int _unsafe = 76;
	const int _ushort = 77;
	const int _usingKW = 78;
	const int _virtual = 79;
	const int _void = 80;
	const int _volatile = 81;
	const int _while = 82;
	const int _dot = 83;
	const int _colon = 84;
	const int _dblcolon = 85;
	const int _comma = 86;
	const int _scolon = 87;
	const int _lbrace = 88;
	const int _rbrace = 89;
	const int _lbrack = 90;
	const int _rbrack = 91;
	const int _lpar = 92;
	const int _rpar = 93;
	const int _tilde = 94;
	const int _condition = 95;
	const int _assgn = 96;
	const int _equal = 97;
	const int _nequal = 98;
	const int _not = 99;
	const int _and = 100;
	const int _andassgn = 101;
	const int _andand = 102;
	const int _or = 103;
	const int _orassgn = 104;
	const int _oror = 105;
	const int _xor = 106;
	const int _xorassgn = 107;
	const int _plus = 108;
	const int _plusassgn = 109;
	const int _inc = 110;
	const int _minus = 111;
	const int _minusassgn = 112;
	const int _dec = 113;
	const int _div = 114;
	const int _divassgn = 115;
	const int _times = 116;
	const int _timesassgn = 117;
	const int _mod = 118;
	const int _modassgn = 119;
	const int _gt = 120;
	const int _gteq = 121;
	const int _lt = 122;
	const int _lteq = 123;
	const int _ltlt = 124;
	const int _lshassgn = 125;
	const int maxT = 129;

	const bool T = true;
	const bool x = false;
	const int minErrDist = 2;
	
	private RootInfo ciRoot;
	private CodeInfo ciCur;
	
	public Scanner scanner;
	public Errors  errors;

	public Token t;    // last recognized token
	public Token la;   // lookahead token
	int errDist = minErrDist;

const int maxTerminals = 160;  // set size

//Statement-Complexity - Gewichte nach Rechenberg
const float rechSC_Assign = 1.0f;
const float rechSC_Method = 1.0f;
const float rechSC_If = 2.0f;
const float rechSC_Else = 1.0f;
const float rechSC_Select = 3.0f;
const float rechSC_While = 3.0f;	//Für While und Do-While Schleifen
const float rechSC_For = 3.0f;	//Für For und Foreach Schleifen
const float rechSC_Return = 1.0f;
const float rechSC_Goto = 5.0f;
const float rechSC_Break = 4.0f;
const float rechSC_Continue = 3.0f;
const float rechSC_Throw = 2.0f;
const float rechSC_Try = 1.0f;
//Statement-Complexity - Verschachtelungsfaktor nach Rechenberg
const float rechSC_Base = 1.5f;

//Expression-Complexity - Gewichte nach Rechenberg
const float rechEC_Method = 1.0f;
const float rechEC_Add = 1.0f;
const float rechEC_Mul = 2.0f;
const float rechEC_Mod = 3.0f;
const float rechEC_Rel = 1.0f;
const float rechEC_Not = 2.0f;
const float rechEC_And = 3.0f;
const float rechEC_Indx = 2.0f;
const float rechEC_Fld = 1.0f;
const float rechSC_Bin = 2.0f;
const float rechEC_Shft = 3.0f;
const float rechEC_Deref = 3.0f;
//Expression-Complexity - Verschachtelungsfaktor nach Rechenberg
const float rechEC_Base = 1.5f;

static BitArray NewSet(params int[] values) {
  BitArray a = new BitArray(maxTerminals);
  foreach (int x in values) a[x] = true;
  return a;
}

public static BitArray
  //Dieses Set wird für die Code-Info's gebraucht
  ci_operator  = NewSet(_tilde, _condition, _assgn, _equal, _nequal, _not, _and, 
						_andand, _or, _orassgn, _oror, _xor, _xorassgn, 
						_plusassgn, _inc, _minus, _minusassgn, _dec, _div, 
						_times, _timesassgn, _mod, _modassgn, _gt, _gteq, 
						_lt, _lteq, _ltlt, _lshassgn),
  //Dieses Set dient zur Unterscheidung zw. Operatoren und Operanden im Sinn von Halstead
  ci_HalsteadOperands =
				 NewSet(/*identifier*/ _ident,
						/*constants*/ _intCon, _realCon, _charCon, _stringCon),
  //Dieses Set dient zum Finden der Operatoren, die nicht gezählt werden (z.B. schließende Klammern),
  //da nur Operatoren-Paare gezählt werden
  ci_HalsteadNotCount = 
				 NewSet(/*do-while is one operant, we count while*/_do,
						/*only count opening parentheses*/ _rbrace, _rbrack, _rpar),
  
  unaryOp      = NewSet(_plus, _minus, _not, _tilde, _inc, _dec, _true, _false),
  typeKW       = NewSet(_char, _bool, _object, _string, _sbyte, _byte, _short,
                 _ushort, _int, _uint, _long, _ulong, _float, _double, _decimal),
  unaryHead    = NewSet(_plus, _minus, _not, _tilde, _times, _inc, _dec, _and),
  assnStartOp  = NewSet(_plus, _minus, _not, _tilde, _times),
  castFollower = NewSet(_and, _ident, _intCon, _realCon, _charCon, _stringCon, _lpar, 
                 _new, _this, _base, _null, _checked, _unchecked, _typeof, _sizeof).Or(unaryOp),
  keyword      = NewSet(_abstract, _as, _base, _bool, _break, _byte, _case, _catch,
                 _char, _checked, _class, _const, _continue, _decimal, _default,
                 _delegate, _do, _double, _else, _enum, _event, _explicit,
                 _extern, _false, _finally, _fixed, _float, _for, _foreach,
                 _goto, _if, _implicit, _in, _int, _interface, _internal,
                 _is, _lock, _long, _namespace, _new, _null, _object, _operator,
                 _out, _override, _params, _private, _protected, _public,
                 _readonly, _ref, _return, _sbyte, _sealed, _short, _sizeof,
                 _stackalloc, _static, _string, _struct, _switch, _this, _throw,
                 _true, _try, _typeof, _uint, _ulong, _unchecked, _unsafe,
                 _ushort, _usingKW, _virtual, _void, _volatile, _while),
  assgnOps     = NewSet(_assgn, _plusassgn, _minusassgn, _timesassgn, _divassgn,
                 _modassgn, _andassgn, _orassgn, _xorassgn, _lshassgn) /* rshassgn: ">" ">="  no whitespace allowed*/
                 ;
 
 //Diese Map wird gebraucht, um die Operatoren ausgeben zu können
 static Hashtable NewOperatorMap() {
	Hashtable htOp = new Hashtable();
	
	htOp[_tilde      ] = "~";
	htOp[_condition  ] = "?";
	htOp[_assgn      ] = "=";
	htOp[_equal      ] = "==";
	htOp[_nequal     ] = "!=";
	htOp[_not        ] = "!";
	htOp[_and        ] = "&";
	htOp[_andassgn   ] = "&=";
	htOp[_andand     ] = "&&";
	htOp[_or         ] = "|";
	htOp[_orassgn    ] = "|=";
	htOp[_oror       ] = "||";  
	htOp[_xor        ] = "^";
	htOp[_xorassgn   ] = "^=";
	htOp[_plus       ] = "+";
	htOp[_plusassgn  ] = "+=";
	htOp[_inc        ] = "++";
	htOp[_minus      ] = "-";
	htOp[_minusassgn ] = "-=";
	htOp[_dec        ] = "--";
	htOp[_div        ] = "/";
	htOp[_divassgn   ] = "/=";
	htOp[_times      ] = "*";
	htOp[_timesassgn ] = "*=";
	htOp[_mod        ] = "%";
	htOp[_modassgn   ] = "%=";
	htOp[_gt         ] = ">";
	htOp[_gteq       ] = ">=";
	htOp[_lt         ] = "<";
	htOp[_lteq       ] = "<=";
	htOp[_ltlt       ] = "<<";
	htOp[_lshassgn   ] = "<<=";

	return htOp;
}

public static Hashtable ci_htOperators = NewOperatorMap();

/*---------------------------- auxiliary methods ------------------------*/

void Error (string s) {
  if (errDist >= minErrDist) errors.Error(la.line, la.col, s);
  errDist = 0;
}

// Return the n-th token after the current lookahead token
Token Peek (int n) {
  scanner.ResetPeek();
  Token x = la;
  while (n > 0) { x = scanner.Peek(); n--; }
  return x;
}

// ident "="
bool IsAssignment () {
  return la.kind == _ident && Peek(1).kind == _assgn;
}

/* True, if the comma is not a trailing one, *
 * like the last one in: a, b, c,            */
bool NotFinalComma () {
  int peek = Peek(1).kind;
  return la.kind == _comma && peek != _rbrace && peek != _rbrack;
}

// NonGenericType ident
// (Type can be void*)
bool IsLocalVarDecl() {
  Token pt;
  
  scanner.ResetPeek();
  if (typeKW[la.kind] || la.kind == _void) {
    pt = scanner.Peek();
    if (la.kind == _void && pt.kind != _times) { return false; }
  } else if (la.kind == _ident) {
    pt = scanner.Peek();
    if (pt.kind == _dblcolon) {
      pt = scanner.Peek();
      if (pt.kind != _ident) {
        return false;
      }
    }
    while (pt.kind == _dot) {
      pt = scanner.Peek();
      if (pt.kind != _ident) { return false; }
      pt = scanner.Peek();
    }
  } else {
    return false;
  }
  
  return SkipPointerOrDims(ref pt) && pt.kind == _ident;
}

// "[" ("," | "]")
bool IsDims () {
  int peek = Peek(1).kind;
  return la.kind == _lbrack && (peek == _comma || peek == _rbrack);
}

// "*" | "[" ("," | "]")
bool IsPointerOrDims () {
  return la.kind == _times || IsDims();
}

bool IsDims(ref Token pt) {
  int peek = pt.kind;
  return pt.kind == _lbrack && ((pt = scanner.Peek()).kind == _comma || pt.kind == _rbrack);
}

bool IsPointerOrDims(ref Token pt) {
  return pt.kind == _times || IsDims(ref pt);
}

/* skip: { "[" { "," } "]" | "*" }             */
/* !!! Proceeds from current peek position !!! */
bool SkipPointerOrDims (ref Token pt) {
  for (;;) {
    if (pt.kind == _lbrack) {
      do pt = scanner.Peek();
      while (pt.kind == _comma);
      if (pt.kind != _rbrack) return false;
    } else if (pt.kind != _times) break;
    pt = scanner.Peek();
  }
  return true;
}

// Is attribute target specifier
// (ident | keyword) ":"
bool IsAttrTargSpec () {
  return (la.kind == _ident || keyword[la.kind]) && Peek(1).kind == _colon;
}

// ident ("," | "=" | ";")
bool IsFieldDecl () {
  int peek = Peek(1).kind;
  return la.kind == _ident && 
         (peek == _comma || peek == _assgn || peek == _scolon);
}

bool IsTypeCast () {
  if (la.kind != _lpar) return false;
  if (IsSimpleTypeCast()) return true;
  return GuessTypeCast();
}

// "(" ( typeKW [ "[" {","} "]" | "*" ] | void  ( "[" {","} "]" | "*" ) ) ")"
bool IsSimpleTypeCast () {
  // assert: la.kind == _lpar
  scanner.ResetPeek();
  Token pt1 = scanner.Peek();
  Token pt = scanner.Peek();
  
  if (typeKW[pt1.kind]) {
    if (pt.kind == _times || pt.kind == _lbrack) {
      return IsPointerOrDims(ref pt) && pt.kind == _rpar;
    } else {
      return pt.kind == _rpar;
    }
  } else if (pt1.kind == _void) {
    return IsPointerOrDims(ref pt) && pt.kind == _rpar;
  }
  
  return false;
}

// "(" NonGenericTypeName { "*" | "[" {","} "]" } ")" castFollower
// NonGenericTypeName = ident [ "::" ident ] {  "." ident }
bool GuessTypeCast () {
  // assert: la.kind == _lpar
  scanner.ResetPeek();
  Token pt = scanner.Peek();
  // ident
  if (pt.kind != _ident && !typeKW[pt.kind]) {
    return false;
  } 
  pt = scanner.Peek();
  // "::" ident
  if (pt.kind == _dblcolon) {
    pt = scanner.Peek();
    if (pt.kind != _ident) {
      return false;
    }
    pt = scanner.Peek();
  }
  // { "." ident }
  while (pt.kind == _dot) {
    pt = scanner.Peek();
    if (pt.kind != _ident) {
      return false;
    }
    pt = scanner.Peek();
  }
  // { "*" | "[" {","} "]" }
  if (!SkipPointerOrDims(ref pt)) {
    return false;
  }
  // ")"
  if (pt.kind != _rpar) {
    return false;
  }
  // check successor
  pt = scanner.Peek();
  return castFollower[pt.kind] || (typeKW[pt.kind] && scanner.Peek().kind == _dot);
}


// "[" "assembly"
bool IsGlobalAttrTarget () {
  Token pt = Peek(1);
  return la.kind == _lbrack && pt.kind == _ident && pt.val == "assembly";
}

// "extern" "alias"
// where alias is an identifier, no keyword
bool IsExternAliasDirective () {
  return la.kind == _extern && Peek(1).val == "alias";
}

// true: anyToken"<"
// no whitespace between the token and the "<" allowed
// anything else will return false.
bool IsLtNoWs() {
  return (la.kind == _lt) && ((t.pos + t.val.Length) == la.pos);
}

bool IsNoSwitchLabelOrRBrace() {
  return (la.kind != _case && la.kind != _default && la.kind != _rbrace) ||
         (la.kind == _default && Peek(1).kind != _colon);
}

bool IsShift() {
  Token pt = Peek(1);
  return (la.kind == _ltlt) ||
         ( la.kind == _gt &&
           pt.kind == _gt &&
           (la.pos + la.val.Length == pt.pos)
         );
}

enum TypeKind {simple, array, pointer, @void}

[Flags]
enum Operator {
  plus   = 0x00000001, minus  = 0x00000002, not    = 0x00000004, tilde  = 0x00000008,
  inc    = 0x00000010, dec    = 0x00000020, @true  = 0x00000040, @false = 0x00000080,
  times  = 0x00000100, div    = 0x00000200, mod    = 0x00000400, and    = 0x00000800,
  or     = 0x00001000, xor    = 0x00002000, lshift = 0x00004000, rshift = 0x00008000,
  eq     = 0x00010000, neq    = 0x00020000, gt     = 0x00040000, lt     = 0x00080000,
  gte    = 0x00100000, lte    = 0x00200000,
  unary  = plus|minus|not|tilde|inc|dec|@true|@false,
  binary = plus|minus|times|div|mod|and|or|xor|lshift|rshift|eq|neq|gt|lt|gte|lte
}

/*------------------------- modifier handling -----------------------------*/

[Flags]
enum Modifier {
  @new      = 0x0001, @public = 0x0002, @protected= 0x0004, @internal = 0x0008,
  @private  = 0x0010, @unsafe = 0x0020, @static   = 0x0040, @readonly = 0x0080,
  @volatile = 0x0100, @virtual= 0x0200, @sealed   = 0x0400, @override = 0x0800,
  @abstract = 0x1000, @extern = 0x2000,
  
  /* sets of modifiers that can be attached to certain program elements    *
   * e.g., "constants" marks all modifiers that may be used with constants */
  none             = 0x0000,
  classes          = @new|@public|@protected|@internal|@private|@unsafe|@abstract|@sealed|@static,
  constants        = @new|@public|@protected|@internal|@private,
  fields           = @new|@public|@protected|@internal|@private|@unsafe|@static|@readonly|@volatile,
  propEvntMeths    = @new|@public|@protected|@internal|@private|@unsafe|@static|@virtual|@sealed|@override|@abstract|@extern,
  accessorsPossib1 = @private,
  accessorsPossib2 = @protected|@internal,
  indexers         = @new|@public|@protected|@internal|@private|@unsafe|@virtual|@sealed|@override|@abstract|@extern,
  operators        = @public|@unsafe|@static|@extern,
  operatorsMust    = @public|@static,
  constructors     = @public|@protected|@internal|@private|@unsafe|@extern,
  staticConstr     = @extern|@static,
  staticConstrMust = @static,
  nonClassTypes    = @new|@public|@protected|@internal|@private|@unsafe,
  destructors      = @extern|@unsafe,
  all              = 0x3fff
}

class Modifiers {
  private Modifier cur = Modifier.none;
  private Parser parser;
  
  public Modifiers(Parser parser) {
  	this.parser = parser;
  }
  
  public void Add (Modifier m) {
    if ((cur & m) == 0) cur |= m;
    else parser.Error("modifier " + m + " already defined");
  }
  
  public void Add (Modifiers m) { Add(m.cur); }

  public bool IsNone { get { return cur == Modifier.none; } }

  public void Check (Modifier allowed) {
    Modifier wrong = cur & (allowed ^ Modifier.all);
    if (wrong != Modifier.none)
      parser.Error("modifier(s) " + wrong + " not allowed here");
  }
  
  public void Check (Modifier allowEither, Modifier allowOr) {
    Modifier wrong = cur & ((allowEither|allowOr) ^ Modifier.all);
    if ((allowEither&allowOr) != Modifier.none) {
      parser.Error("modifiers providerd must not overlap");
    } else if (wrong != Modifier.none) {
      parser.Error("modifier(s) " + wrong + " not allowed here");
    } else if (((cur&allowEither) != Modifier.none) && ((cur&allowOr) != Modifier.none)) {
      parser.Error("modifier(s) may either be " + allowEither + " or " + allowOr);
    }
  }
  
  public void CheckMust (Modifier mustHave) {
    Modifier missing = (cur&mustHave)^mustHave;
    if (missing != Modifier.none) {
      parser.Error("modifier(s) " + missing + " must be applied here");
    }
  }
  
  public bool Has (Modifier mod) {
    return (cur&mod) == mod;
  }
}


/*------------------------------------------------------------------------*
 *----- SCANNER DESCRIPTION ----------------------------------------------*
 *------------------------------------------------------------------------*/



	public Parser(RootInfo ciRoot, Scanner scanner) {
		this.ciRoot = ciRoot;
		this.ciCur = ciRoot;
		this.scanner = scanner;
		scanner.setCurCodeInfo(ciCur);
		errors = new Errors();
	}

	void ChangeCurCodeInfo(CodeInfo ciCur) {
		this.ciCur = ciCur;
		scanner.setCurCodeInfo(ciCur);
	}
	
	void ChangeToParentCodeInfo() {
		if (ciCur.getParent() != null) {
			ciCur = ciCur.getParent();
			scanner.setCurCodeInfo(ciCur);
		}
		else {
			Console.WriteLine("Internal Error: more closed regions as opened!");
		}
	}

	void SynErr (int n) {
		if (errDist >= minErrDist) errors.SynErr(la.line, la.col, n);
		errDist = 0;
	}

	public void SemErr (string msg) {
		if (errDist >= minErrDist) errors.Error(t.line, t.col, msg);
		errDist = 0;
	}
	
	void Get () {
		for (;;) {
			t = la;
			la = scanner.Scan();
			if (la.kind <= maxT) { ++errDist; break; }
				if (la.kind == 130) {
				}
				if (la.kind == 131) {
				}
				if (la.kind == 132) {
				}
				if (la.kind == 133) {
				}
				if (la.kind == 134) {
				}
				if (la.kind == 135) {
				}
				if (la.kind == 136) {
				}
				if (la.kind == 137) {
				}
				if (la.kind == 138) {
				}
				if (la.kind == 139) {
				}
				if (la.kind == 140) {
				}

			la = t;
		}
	}
	
	void Expect (int n) {
		if (la.kind==n) Get(); else { SynErr(n); }
	}
	
	bool StartOf (int s) {
		return set[s, la.kind];
	}
	
	void ExpectWeak (int n, int follow) {
		if (la.kind == n) Get();
		else {
			SynErr(n);
			while (!StartOf(follow)) Get();
		}
	}
	
	bool WeakSeparator (int n, int syFol, int repFol) {
		bool[] s = new bool[maxT+1];
		if (la.kind == n) { Get(); return true; }
		else if (StartOf(repFol)) return false;
		else {
			for (int i=0; i <= maxT; i++) {
				s[i] = set[syFol, i] || set[repFol, i] || set[0, i];
			}
			SynErr(n);
			while (!s[la.kind]) Get();
			return StartOf(syFol);
		}
	}
	
	void CS2() {
		while (IsExternAliasDirective()) {
			ExternAliasDirective();
		}
		while (la.kind == 78) {
			UsingDirective();
		}
		while (IsGlobalAttrTarget()) {
			GlobalAttributes();
		}
		while (StartOf(1)) {
			NamespaceMemberDeclaration();
		}
	}

	void ExternAliasDirective() {
		Expect(28);
		Expect(1);
		if (t.val != "alias") {
		 Error("alias expected");
		}
		
		Expect(1);
		Expect(87);
	}

	void UsingDirective() {
		String tname; bool isAlias = false; 
		String name = "";
		
		Expect(78);
		if (IsAssignment()) {
			Expect(1);
			isAlias = true; name = t.val; 
			Expect(96);
		}
		TypeName(out tname);
		if (ciCur.GetType() == typeof(RootInfo)) {
		if (isAlias)
			((RootInfo)ciCur).uiUsings.addAlias(name, tname); 
		else
			((RootInfo)ciCur).uiUsings.addUsing(tname);
		}
		else {
			if (isAlias)
				((NamespaceInfo)ciCur).uiUsings.addAlias(name, tname);
			else
				((NamespaceInfo)ciCur).uiUsings.addUsing(tname);
		}		
		
		Expect(87);
	}

	void GlobalAttributes() {
		Expect(90);
		Expect(1);
		if (t.val != "assembly") Error("global attribute target specifier \"assembly\" expected"); 
		Expect(84);
		Attribute();
		while (NotFinalComma()) {
			Expect(86);
			Attribute();
		}
		if (la.kind == 86) {
			Get();
		}
		Expect(91);
	}

	void NamespaceMemberDeclaration() {
		Modifiers m = new Modifiers(this); 
		if (la.kind == 45) {
			Get();
			Expect(1);
			int iNsLine = t.line;
			StringBuilder sb = new StringBuilder();
			sb.Append(t.val);
			
			while (la.kind == 83) {
				Get();
				Expect(1);
				sb.Append(".").Append(t.val); 
			}
			NamespaceInfo nsInfo = new NamespaceInfo(sb.ToString(), iNsLine, ciCur);
			ChangeCurCodeInfo(nsInfo);
			
			Expect(88);
			while (IsExternAliasDirective()) {
				ExternAliasDirective();
			}
			while (la.kind == 78) {
				UsingDirective();
			}
			while (StartOf(1)) {
				NamespaceMemberDeclaration();
			}
			Expect(89);
			ChangeToParentCodeInfo(); 
			if (la.kind == 87) {
				Get();
			}
		} else if (StartOf(2)) {
			while (la.kind == 90) {
				Attributes();
			}
			ModifierList(m);
			TypeDeclaration();
		} else SynErr(130);
	}

	void TypeName(out String name) {
		StringBuilder sb = new StringBuilder();
		String tname;
		
		Expect(1);
		sb.Append(t.val); 
		if (la.kind == 85) {
			Get();
			Expect(1);
			sb.Append("::").Append(t.val); 
		}
		if (la.kind == 122) {
			TypeArgumentList(out tname);
		}
		while (la.kind == 83) {
			Get();
			Expect(1);
			sb.Append(".").Append(t.val); 
			if (la.kind == 122) {
				TypeArgumentList(out tname);
			}
		}
		name = sb.ToString(); 
	}

	void Attributes() {
		Expect(90);
		if (IsAttrTargSpec()) {
			if (la.kind == 1) {
				Get();
			} else if (StartOf(3)) {
				Keyword();
			} else SynErr(131);
			Expect(84);
		}
		Attribute();
		while (la.kind == _comma && Peek(1).kind != _rbrack) {
			Expect(86);
			Attribute();
		}
		if (la.kind == 86) {
			Get();
		}
		Expect(91);
	}

	void ModifierList(Modifiers m) {
		while (StartOf(4)) {
			switch (la.kind) {
			case 46: {
				Get();
				m.Add(Modifier.@new); 
				break;
			}
			case 55: {
				Get();
				m.Add(Modifier.@public); 
				break;
			}
			case 54: {
				Get();
				m.Add(Modifier.@protected); 
				break;
			}
			case 41: {
				Get();
				m.Add(Modifier.@internal); 
				break;
			}
			case 53: {
				Get();
				m.Add(Modifier.@private); 
				break;
			}
			case 76: {
				Get();
				m.Add(Modifier.@unsafe); 
				break;
			}
			case 64: {
				Get();
				m.Add(Modifier.@static); 
				break;
			}
			case 56: {
				Get();
				m.Add(Modifier.@readonly); 
				break;
			}
			case 81: {
				Get();
				m.Add(Modifier.@volatile); 
				break;
			}
			case 79: {
				Get();
				m.Add(Modifier.@virtual); 
				break;
			}
			case 60: {
				Get();
				m.Add(Modifier.@sealed); 
				break;
			}
			case 51: {
				Get();
				m.Add(Modifier.@override); 
				break;
			}
			case 6: {
				Get();
				m.Add(Modifier.@abstract); 
				break;
			}
			case 28: {
				Get();
				m.Add(Modifier.@extern); 
				break;
			}
			}
		}
	}

	void TypeDeclaration() {
		Modifiers m = new Modifiers(this);
		TypeKind dummy; String tname;
		
		if (StartOf(5)) {
			if (la.kind == 126) {
				Get();
			}
			if (la.kind == 16) {
				m.Check(Modifier.classes); 
				Get();
				Expect(1);
				ClassInfo csInfo = new ClassInfo(t.val, t.line, ciCur);
				ChangeCurCodeInfo(csInfo);
				
				if (la.kind == 122) {
					TypeParameterList();
				}
				if (la.kind == 84) {
					ClassBase();
				}
				while (la.kind == 1) {
					TypeParameterConstraintsClause();
				}
				ClassBody();
				if (la.kind == 87) {
					Get();
				}
			} else if (la.kind == 66) {
				m.Check(Modifier.nonClassTypes); 
				Get();
				Expect(1);
				ClassInfo csInfo = new ClassInfo(t.val, t.line, ciCur);
				ChangeCurCodeInfo(csInfo);
				
				if (la.kind == 122) {
					TypeParameterList();
				}
				if (la.kind == 84) {
					Get();
					TypeName(out tname);
					while (la.kind == 86) {
						Get();
						TypeName(out tname);
					}
				}
				while (la.kind == 1) {
					TypeParameterConstraintsClause();
				}
				StructBody();
				if (la.kind == 87) {
					Get();
				}
			} else if (la.kind == 40) {
				m.Check(Modifier.nonClassTypes); 
				Get();
				Expect(1);
				ClassInfo csInfo = new ClassInfo(t.val, t.line, ciCur);
				ChangeCurCodeInfo(csInfo);
				
				if (la.kind == 122) {
					TypeParameterList();
				}
				if (la.kind == 84) {
					Get();
					TypeName(out tname);
					while (la.kind == 86) {
						Get();
						TypeName(out tname);
					}
				}
				while (la.kind == 1) {
					TypeParameterConstraintsClause();
				}
				Expect(88);
				while (StartOf(6)) {
					InterfaceMemberDeclaration();
				}
				Expect(89);
				if (la.kind == 87) {
					Get();
				}
			} else SynErr(132);
			ChangeToParentCodeInfo(); 
		} else if (la.kind == 25) {
			m.Check(Modifier.nonClassTypes); 
			Get();
			Expect(1);
			if (la.kind == 84) {
				Get();
				IntegralType(out tname);
			}
			EnumBody();
			if (la.kind == 87) {
				Get();
			}
		} else if (la.kind == 21) {
			m.Check(Modifier.nonClassTypes); 
			Get();
			Type(out dummy, true, out tname);
			Expect(1);
			MethodInfo mInfo = new MethodInfo(t.val, t.line, ciCur);
			ChangeCurCodeInfo(mInfo);
			
			if (la.kind == 122) {
				TypeParameterList();
			}
			Expect(92);
			if (StartOf(7)) {
				FormalParameterList();
			}
			Expect(93);
			while (la.kind == 1) {
				TypeParameterConstraintsClause();
			}
			Expect(87);
			ChangeToParentCodeInfo(); 
		} else SynErr(133);
	}

	void TypeParameterList() {
		Expect(122);
		while (la.kind == 90) {
			Attributes();
		}
		Expect(1);
		while (la.kind == 86) {
			Get();
			while (la.kind == 90) {
				Attributes();
			}
			Expect(1);
		}
		Expect(120);
	}

	void ClassBase() {
		String tname; 
		Expect(84);
		ClassType(out tname);
		((ClassInfo) ciCur).parentClass = tname; 
		while (la.kind == 86) {
			Get();
			TypeName(out tname);
		}
	}

	void TypeParameterConstraintsClause() {
		String tname; 
		Expect(1);
		if (t.val != "where") {
		 Error("type parameter constraints clause must start with: where");
		}
		
		Expect(1);
		Expect(84);
		if (StartOf(8)) {
			if (la.kind == 16) {
				Get();
			} else if (la.kind == 66) {
				Get();
			} else if (la.kind == 48) {
				Get();
			} else if (la.kind == 65) {
				Get();
			} else {
				TypeName(out tname);
			}
			while (la.kind == _comma && Peek(1).kind != _new) {
				Expect(86);
				TypeName(out tname);
			}
			if (la.kind == 86) {
				Get();
				Expect(46);
				Expect(92);
				Expect(93);
			}
		} else if (la.kind == 46) {
			Get();
			Expect(92);
			Expect(93);
		} else SynErr(134);
	}

	void ClassBody() {
		Expect(88);
		while (StartOf(9)) {
			while (la.kind == 90) {
				Attributes();
			}
			Modifiers m = new Modifiers(this); 
			ModifierList(m);
			ClassMemberDeclaration(m);
		}
		Expect(89);
	}

	void StructBody() {
		Expect(88);
		while (StartOf(10)) {
			while (la.kind == 90) {
				Attributes();
			}
			Modifiers m = new Modifiers(this); 
			ModifierList(m);
			StructMemberDeclaration(m);
		}
		Expect(89);
	}

	void InterfaceMemberDeclaration() {
		Modifiers m = new Modifiers(this);
		TypeKind dummy; String tname;
		
		while (la.kind == 90) {
			Attributes();
		}
		if (la.kind == 46) {
			Get();
		}
		if (StartOf(11)) {
			Type(out dummy, true, out tname);
			if (la.kind == 1) {
				Get();
				MethodInfo mInfo = new MethodInfo(t.val, t.line, ciCur);
				ChangeCurCodeInfo(mInfo);
				
				if (la.kind == 92 || la.kind == 122) {
					if (la.kind == 122) {
						TypeParameterList();
					}
					Expect(92);
					if (StartOf(7)) {
						FormalParameterList();
					}
					Expect(93);
					while (la.kind == 1) {
						TypeParameterConstraintsClause();
					}
					Expect(87);
				} else if (la.kind == 88) {
					Get();
					InterfaceAccessors(m);
					Expect(89);
				} else SynErr(135);
				ChangeToParentCodeInfo(); 
			} else if (la.kind == 68) {
				Get();
				MethodInfo mInfo = new MethodInfo("this[]", t.line, ciCur);
				ChangeCurCodeInfo(mInfo);
				
				Expect(90);
				FormalParameterList();
				Expect(91);
				Expect(88);
				InterfaceAccessors(m);
				Expect(89);
				ChangeToParentCodeInfo(); 
			} else SynErr(136);
		} else if (la.kind == 26) {
			Get();
			Type(out dummy, false, out tname);
			Expect(1);
			Expect(87);
		} else SynErr(137);
	}

	void IntegralType(out String name) {
		name = ""; 
		switch (la.kind) {
		case 59: {
			Get();
			name = "sbyte"; 
			break;
		}
		case 11: {
			Get();
			name = "byte"; 
			break;
		}
		case 61: {
			Get();
			name = "short"; 
			break;
		}
		case 77: {
			Get();
			name = "ushort"; 
			break;
		}
		case 39: {
			Get();
			name = "int"; 
			break;
		}
		case 73: {
			Get();
			name = "uint"; 
			break;
		}
		case 44: {
			Get();
			name = "long"; 
			break;
		}
		case 74: {
			Get();
			name = "ulong"; 
			break;
		}
		case 14: {
			Get();
			name = "char"; 
			break;
		}
		default: SynErr(138); break;
		}
	}

	void EnumBody() {
		Expect(88);
		if (la.kind == 1 || la.kind == 90) {
			EnumMemberDeclaration();
			while (NotFinalComma()) {
				Expect(86);
				EnumMemberDeclaration();
			}
			if (la.kind == 86) {
				Get();
			}
		}
		Expect(89);
	}

	void Type(out TypeKind type, bool voidAllowed, out String name) {
		type = TypeKind.simple; 
		name = "";
		
		if (StartOf(12)) {
			PrimitiveType(out name);
		} else if (la.kind == 1 || la.kind == 48 || la.kind == 65) {
			ClassType(out name);
		} else if (la.kind == 80) {
			Get();
			type = TypeKind.@void;
			name = "void";
			
		} else SynErr(139);
		PointerOrArray(ref type);
		if (type == TypeKind.@void && !voidAllowed) Error("type expected, void found, maybe you mean void*"); 
	}

	void FormalParameterList() {
		TypeKind type; String tname; 
		while (la.kind == 90) {
			Attributes();
		}
		if (StartOf(13)) {
			if (la.kind == 50 || la.kind == 57) {
				if (la.kind == 57) {
					Get();
				} else {
					Get();
				}
			}
			Type(out type, false, out tname);
			Expect(1);
			VariableInfo vInfo = new VariableInfo(t.val, tname, ciCur);
			((MethodInfo) ciCur).addParameter(vInfo);
			
			if (la.kind == 86) {
				Get();
				FormalParameterList();
			}
		} else if (la.kind == 52) {
			Get();
			Type(out type, false, out tname);
			if (type != TypeKind.array) { Error("params argument must be an array"); } 
			Expect(1);
			VariableInfo vInfo = new VariableInfo(t.val, tname, ciCur);
			((MethodInfo) ciCur).addParameter(vInfo);
			
		} else SynErr(140);
	}

	void ClassType(out String name) {
		name = ""; 
		if (la.kind == 1) {
			TypeName(out name);
		} else if (la.kind == 48 || la.kind == 65) {
			InternalClassType(out name);
		} else SynErr(141);
	}

	void ClassMemberDeclaration(Modifiers m) {
		if (StartOf(14)) {
			StructMemberDeclaration(m);
		} else if (la.kind == 94) {
			Get();
			Expect(1);
			MethodInfo mInfo = new MethodInfo("~" + t.val, t.line, ciCur);
			ChangeCurCodeInfo(mInfo);
			
			Expect(92);
			Expect(93);
			if (la.kind == 88) {
				Block();
			} else if (la.kind == 87) {
				Get();
			} else SynErr(142);
			ChangeToParentCodeInfo(); 
		} else SynErr(143);
	}

	void StructMemberDeclaration(Modifiers m) {
		TypeKind type; Operator op;
		String tname; Hashtable htTmp = new Hashtable();
		
		if (la.kind == 17) {
			m.Check(Modifier.constants); 
			Get();
			Type(out type, false, out tname);
			Expect(1);
			VariableInfo vInfo = new VariableInfo(t.val, tname, ciCur);
			ciCur.addVariable(vInfo);
			
			Expect(96);
			Expression(ref htTmp);
			while (la.kind == 86) {
				Get();
				Expect(1);
				vInfo = new VariableInfo(t.val, tname, ciCur);
				ciCur.addVariable(vInfo);
				
				Expect(96);
				Expression(ref htTmp);
			}
			Expect(87);
		} else if (la.kind == 26) {
			m.Check(Modifier.propEvntMeths); 
			Get();
			Type(out type, false, out tname);
			if (IsFieldDecl()) {
				VariableDeclarators(tname);
				Expect(87);
			} else if (la.kind == 1) {
				TypeName(out tname);
				Expect(88);
				EventAccessorDeclarations();
				Expect(89);
			} else SynErr(144);
		} else if (la.kind == _ident && Peek(1).kind == _lpar) {
			m.Check(Modifier.constructors|Modifier.staticConstr); 
			Expect(1);
			MethodInfo mInfo = new MethodInfo(t.val, t.line, ciCur);
			ChangeCurCodeInfo(mInfo);
			
			Expect(92);
			if (StartOf(7)) {
				m.Check(Modifier.constructors); 
				FormalParameterList();
			}
			Expect(93);
			if (la.kind == 84) {
				m.Check(Modifier.constructors); 
				Get();
				if (la.kind == 8) {
					Get();
				} else if (la.kind == 68) {
					Get();
				} else SynErr(145);
				Expect(92);
				if (StartOf(15)) {
					Argument();
					while (la.kind == 86) {
						Get();
						Argument();
					}
				}
				Expect(93);
			}
			if (la.kind == 88) {
				Block();
			} else if (la.kind == 87) {
				Get();
			} else SynErr(146);
			ChangeToParentCodeInfo(); 
		} else if (StartOf(11)) {
			Type(out type, true, out tname);
			if (la.kind == 49) {
				m.Check(Modifier.operators);
				m.CheckMust(Modifier.operatorsMust);
				if (type == TypeKind.@void) { Error("operator not allowed on void"); } 
				
				Get();
				OverloadableOp(out op);
				MethodInfo mInfo = null;
				if (op == Operator.rshift)
					mInfo = new MethodInfo("operator>>", t.line, ciCur);
				else
					mInfo = new MethodInfo("operator"+t.val, t.line, ciCur);
				ChangeCurCodeInfo(mInfo);
				
				Expect(92);
				Type(out type, false, out tname);
				Expect(1);
				VariableInfo vInfo = new VariableInfo(t.val, tname, ciCur);
				((MethodInfo) ciCur).addParameter(vInfo);
				
				if (la.kind == 86) {
					Get();
					Type(out type, false, out tname);
					Expect(1);
					if ((op & Operator.binary) == 0) Error("too many operands for unary operator");
					vInfo = new VariableInfo(t.val, tname, ciCur);
					((MethodInfo) ciCur).addParameter(vInfo);
					
				} else if (la.kind == 93) {
					if ((op & Operator.unary) == 0) Error("too few operands for binary operator"); 
				} else SynErr(147);
				Expect(93);
				if (la.kind == 88) {
					Block();
				} else if (la.kind == 87) {
					Get();
				} else SynErr(148);
				ChangeToParentCodeInfo(); 
			} else if (IsFieldDecl()) {
				m.Check(Modifier.fields);
				if (type == TypeKind.@void) { Error("field type must not be void"); } 
				
				VariableDeclarators(tname);
				Expect(87);
			} else if (la.kind == 1) {
				NonGenericTypeName(out tname);
				if (la.kind == 88) {
					m.Check(Modifier.propEvntMeths);
					if (type == TypeKind.@void) { Error("property type must not be void"); }
					MethodInfo mInfo = new MethodInfo(t.val, t.line, ciCur);
					ChangeCurCodeInfo(mInfo);
					
					Get();
					AccessorDeclarations(m);
					Expect(89);
					ChangeToParentCodeInfo(); 
				} else if (la.kind == 83) {
					m.Check(Modifier.indexers);
					if (type == TypeKind.@void) { Error("indexer type must not be void"); } 
					MethodInfo mInfo = new MethodInfo(t.val+"[]", t.line, ciCur);
					ChangeCurCodeInfo(mInfo);
					
					Get();
					Expect(68);
					Expect(90);
					FormalParameterList();
					Expect(91);
					Expect(88);
					AccessorDeclarations(m);
					Expect(89);
					ChangeToParentCodeInfo(); 
				} else if (la.kind == 92 || la.kind == 122) {
					m.Check(Modifier.propEvntMeths);
					MethodInfo mInfo = new MethodInfo(t.val, t.line, ciCur);
					ChangeCurCodeInfo(mInfo);
					
					if (la.kind == 122) {
						TypeParameterList();
					}
					Expect(92);
					if (StartOf(7)) {
						FormalParameterList();
					}
					Expect(93);
					while (la.kind == 1) {
						TypeParameterConstraintsClause();
					}
					if (la.kind == 88) {
						Block();
					} else if (la.kind == 87) {
						Get();
					} else SynErr(149);
					ChangeToParentCodeInfo(); 
				} else SynErr(150);
			} else if (la.kind == 68) {
				m.Check(Modifier.indexers);
				if (type == TypeKind.@void) { Error("indexer type must not be void"); }
				MethodInfo mInfo = new MethodInfo("this[]", t.line, ciCur);
				ChangeCurCodeInfo(mInfo);
				
				Get();
				Expect(90);
				FormalParameterList();
				Expect(91);
				Expect(88);
				AccessorDeclarations(m);
				Expect(89);
				ChangeToParentCodeInfo(); 
			} else SynErr(151);
		} else if (la.kind == 27 || la.kind == 37) {
			m.Check(Modifier.operators);
			m.CheckMust(Modifier.operatorsMust);
			
			if (la.kind == 37) {
				Get();
			} else {
				Get();
			}
			Expect(49);
			Type(out type, false, out tname);
			if (type == TypeKind.@void) { Error("cast type must not be void"); } 
			MethodInfo mInfo = new MethodInfo(t.val+" cast", t.line, ciCur);
			ChangeCurCodeInfo(mInfo);
			
			Expect(92);
			Type(out type, false, out tname);
			Expect(1);
			VariableInfo vInfo = new VariableInfo(t.val, tname, ciCur);
			((MethodInfo) ciCur).addParameter(vInfo);
			
			Expect(93);
			if (la.kind == 88) {
				Block();
			} else if (la.kind == 87) {
				Get();
			} else SynErr(152);
			ChangeToParentCodeInfo(); 
		} else if (StartOf(16)) {
			TypeDeclaration();
		} else SynErr(153);
	}

	void EnumMemberDeclaration() {
		Hashtable htTmp = new Hashtable(); 
		while (la.kind == 90) {
			Attributes();
		}
		Expect(1);
		if (la.kind == 96) {
			Get();
			Expression(ref htTmp);
		}
	}

	void Block() {
		Expect(88);
		while (StartOf(17)) {
			Statement();
		}
		Expect(89);
	}

	void Expression(ref Hashtable htVarAcc) {
		Hashtable htTmp = new Hashtable(); 
		MethodInfo mInfo = ciCur.GetType() == typeof(MethodInfo) ? (MethodInfo) ciCur : null;
		if (mInfo!=null) mInfo.iExpDepth++;
		
		Unary(ref htTmp);
		if (assgnOps[la.kind] || (la.kind == _gt && Peek(1).kind == _gteq)) {
			AssignmentOperator();
			if (mInfo!=null) {
			if (t.val != "=") //keine reine Zuweisung (also +=, -= etc.)
				mInfo.addVarAcc(htTmp, true);
			mInfo.addVarAcc(htTmp, false);
			htTmp = new Hashtable();
			mInfo.fSC += rechSC_Assign*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth);
			}
			
			Expression(ref htTmp);
		} else if (StartOf(18)) {
			if (mInfo!=null) mInfo.addVarAcc(htTmp, true); 
			OrExpr(ref htTmp);
			if (la.kind == 95) {
				Get();
				Expression(ref htTmp);
				Expect(84);
				Expression(ref htTmp);
			}
		} else SynErr(154);
		if (mInfo!=null) {
		mInfo.addVarAcc(htTmp, true);
		foreach(DictionaryEntry de in htTmp)
			htVarAcc[de.Key] = de.Value;
		mInfo.iExpDepth--;
		}
		
	}

	void VariableDeclarators(String tname) {
		Expect(1);
		VariableInfo vInfo = new VariableInfo(t.val, tname, ciCur);
		ciCur.addVariable(vInfo);
		
		if (la.kind == 96) {
			Get();
			VariableInitializer();
		}
		while (la.kind == 86) {
			Get();
			Expect(1);
			vInfo = new VariableInfo(t.val, tname, ciCur);
			ciCur.addVariable(vInfo);
			
			if (la.kind == 96) {
				Get();
				VariableInitializer();
			}
		}
	}

	void EventAccessorDeclarations() {
		bool addFound = false, remFound = false; 
		while (la.kind == 90) {
			Attributes();
		}
		if (la.val == "add") {
			Expect(1);
			addFound = true; 
		} else if (la.val == "remove") {
			Expect(1);
			remFound = true; 
		} else if (la.kind == 1) {
			Get();
			Error("add or remove expected"); 
		} else SynErr(155);
		Block();
		if (la.kind == 1 || la.kind == 90) {
			while (la.kind == 90) {
				Attributes();
			}
			if (la.val == "add") {
				Expect(1);
				if (addFound) Error("add already declared");    
			} else if (la.val == "remove") {
				Expect(1);
				if (remFound) Error("remove already declared"); 
			} else if (la.kind == 1) {
				Get();
				Error("add or remove expected"); 
			} else SynErr(156);
			Block();
		}
	}

	void Argument() {
		Hashtable htTmp = new Hashtable(); 
		if (la.kind == 50 || la.kind == 57) {
			if (la.kind == 57) {
				Get();
			} else {
				Get();
			}
		}
		Expression(ref htTmp);
		if (ciCur.GetType() == typeof(MethodInfo))
		((MethodInfo)ciCur).addVarAcc(htTmp, true);
		
	}

	void OverloadableOp(out Operator op) {
		op = Operator.plus; 
		switch (la.kind) {
		case 108: {
			Get();
			break;
		}
		case 111: {
			Get();
			op = Operator.minus; 
			break;
		}
		case 99: {
			Get();
			op = Operator.not; 
			break;
		}
		case 94: {
			Get();
			op = Operator.tilde; 
			break;
		}
		case 110: {
			Get();
			op = Operator.inc; 
			break;
		}
		case 113: {
			Get();
			op = Operator.dec; 
			break;
		}
		case 70: {
			Get();
			op = Operator.@true; 
			break;
		}
		case 29: {
			Get();
			op = Operator.@false; 
			break;
		}
		case 116: {
			Get();
			op = Operator.times; 
			break;
		}
		case 114: {
			Get();
			op = Operator.div; 
			break;
		}
		case 118: {
			Get();
			op = Operator.mod; 
			break;
		}
		case 100: {
			Get();
			op = Operator.and; 
			break;
		}
		case 103: {
			Get();
			op = Operator.or; 
			break;
		}
		case 106: {
			Get();
			op = Operator.xor; 
			break;
		}
		case 124: {
			Get();
			op = Operator.lshift; 
			break;
		}
		case 97: {
			Get();
			op = Operator.eq; 
			break;
		}
		case 98: {
			Get();
			op = Operator.neq; 
			break;
		}
		case 120: {
			Get();
			op = Operator.gt; 
			if (la.kind == 120) {
				if (la.pos > t.pos+1) Error("no whitespace allowed in right shift operator"); 
				Get();
				op = Operator.rshift;
				//Es wurde ein Right Shift gefunden -> Statements Info's anpassen!
				int i = (int) ciCur.statInfo.htOperators[">"];
				ciCur.statInfo.htOperators[">"] = i - 2;
				if (ciCur.statInfo.htOperators.ContainsKey(">>")) {
					i = (int) ciCur.statInfo.htOperators[">>"];
					ciCur.statInfo.htOperators[">>"] = i + 1;
				}
				else ciCur.statInfo.htOperators[">>"] = 1;
				
			}
			break;
		}
		case 122: {
			Get();
			op = Operator.lt; 
			break;
		}
		case 121: {
			Get();
			op = Operator.gte; 
			break;
		}
		case 123: {
			Get();
			op = Operator.lte; 
			break;
		}
		default: SynErr(157); break;
		}
	}

	void NonGenericTypeName(out String name) {
		StringBuilder sb = new StringBuilder(); 
		Expect(1);
		sb.Append(t.val); 
		if (la.kind == 85) {
			Get();
			Expect(1);
			sb.Append("::").Append(t.val); 
		}
		while (la.kind == _dot && Peek(1).kind == _ident) {
			Expect(83);
			Expect(1);
			sb.Append(".").Append(t.val); 
		}
		name = sb.ToString(); 
	}

	void AccessorDeclarations(Modifiers m) {
		Modifiers am = new Modifiers(this);
		bool getFound = false, setFound = false;
		
		while (la.kind == 90) {
			Attributes();
		}
		ModifierList(am);
		am.Check(Modifier.accessorsPossib1, Modifier.accessorsPossib2); 
		if (la.val == "get") {
			Expect(1);
			getFound = true; 
		} else if (la.val == "set") {
			Expect(1);
			setFound = true; 
		} else if (la.kind == 1) {
			Get();
			Error("set or get expected"); 
		} else SynErr(158);
		if (la.kind == 88) {
			Block();
		} else if (la.kind == 87) {
			Get();
		} else SynErr(159);
		if (StartOf(19)) {
			am = new Modifiers(this); 
			while (la.kind == 90) {
				Attributes();
			}
			ModifierList(am);
			am.Check(Modifier.accessorsPossib1, Modifier.accessorsPossib2); 
			if (la.val == "get") {
				Expect(1);
				if (getFound) Error("get already declared");  
			} else if (la.val == "set") {
				Expect(1);
				if (setFound) Error("set already declared");  
			} else if (la.kind == 1) {
				Get();
				Error("set or get expected"); 
			} else SynErr(160);
			if (la.kind == 88) {
				Block();
			} else if (la.kind == 87) {
				Get();
			} else SynErr(161);
		}
	}

	void InterfaceAccessors(Modifiers m) {
		bool getFound = false, setFound = false; 
		while (la.kind == 90) {
			Attributes();
		}
		if (la.val == "get") {
			Expect(1);
			getFound = true; 
		} else if (la.val == "set") {
			Expect(1);
			setFound = true; 
		} else if (la.kind == 1) {
			Get();
			Error("set or get expected"); 
		} else SynErr(162);
		Expect(87);
		if (la.kind == 1 || la.kind == 90) {
			while (la.kind == 90) {
				Attributes();
			}
			if (la.val == "get") {
				Expect(1);
				if (getFound) Error("get already declared");  
			} else if (la.val == "set") {
				Expect(1);
				if (setFound) Error("set already declared");  
			} else if (la.kind == 1) {
				Get();
				Error("set or get expected"); 
			} else SynErr(163);
			Expect(87);
		}
	}

	void LocalVariableDeclaration() {
		TypeKind dummy; String tname; 
		NonGenericType(out dummy, false, out tname);
		LocalVariableDeclarator(tname);
		while (la.kind == 86) {
			Get();
			LocalVariableDeclarator(tname);
		}
	}

	void NonGenericType(out TypeKind type, bool voidAllowed, out String name) {
		type = TypeKind.simple;
		name = "";
		
		if (StartOf(12)) {
			PrimitiveType(out name);
		} else if (la.kind == 1 || la.kind == 48 || la.kind == 65) {
			NonGenericClassType(out name);
		} else if (la.kind == 80) {
			Get();
			type = TypeKind.@void;
			name = "void";
			
		} else SynErr(164);
		PointerOrArray(ref type);
		if (type == TypeKind.@void && !voidAllowed) Error("type expected, void found, maybe you mean void*"); 
	}

	void LocalVariableDeclarator(String tname) {
		TypeKind dummy; String tname1; 
		Hashtable htTmp = new Hashtable();
		MethodInfo mInfo = ciCur.GetType() == typeof(MethodInfo) ? (MethodInfo)ciCur : null;
		
		Expect(1);
		VariableInfo vInfo = new VariableInfo(t.val, tname, ciCur);
		ciCur.addVariable(vInfo);
		
		if (la.kind == 96) {
			Get();
			if (mInfo!=null) mInfo.fSC += rechSC_Assign*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth); 
			if (StartOf(20)) {
				VariableInitializer();
			} else if (la.kind == 63) {
				Get();
				Type(out dummy, false, out tname1);
				Expect(90);
				Expression(ref htTmp);
				Expect(91);
			} else SynErr(165);
		}
	}

	void VariableInitializer() {
		Hashtable htTmp = new Hashtable(); 
		if (StartOf(21)) {
			Expression(ref htTmp);
		} else if (la.kind == 88) {
			ArrayInitializer();
		} else SynErr(166);
	}

	void ArrayInitializer() {
		Expect(88);
		if (StartOf(20)) {
			VariableInitializer();
			while (NotFinalComma()) {
				Expect(86);
				VariableInitializer();
			}
			if (la.kind == 86) {
				Get();
			}
		}
		Expect(89);
	}

	void Attribute() {
		String tname; 
		TypeName(out tname);
		if (la.kind == 92) {
			AttributeArguments();
		}
	}

	void Keyword() {
		switch (la.kind) {
		case 6: {
			Get();
			break;
		}
		case 7: {
			Get();
			break;
		}
		case 8: {
			Get();
			break;
		}
		case 9: {
			Get();
			break;
		}
		case 10: {
			Get();
			break;
		}
		case 11: {
			Get();
			break;
		}
		case 12: {
			Get();
			break;
		}
		case 13: {
			Get();
			break;
		}
		case 14: {
			Get();
			break;
		}
		case 15: {
			Get();
			break;
		}
		case 16: {
			Get();
			break;
		}
		case 17: {
			Get();
			break;
		}
		case 18: {
			Get();
			break;
		}
		case 19: {
			Get();
			break;
		}
		case 20: {
			Get();
			break;
		}
		case 21: {
			Get();
			break;
		}
		case 22: {
			Get();
			break;
		}
		case 23: {
			Get();
			break;
		}
		case 24: {
			Get();
			break;
		}
		case 25: {
			Get();
			break;
		}
		case 26: {
			Get();
			break;
		}
		case 27: {
			Get();
			break;
		}
		case 28: {
			Get();
			break;
		}
		case 29: {
			Get();
			break;
		}
		case 30: {
			Get();
			break;
		}
		case 31: {
			Get();
			break;
		}
		case 32: {
			Get();
			break;
		}
		case 33: {
			Get();
			break;
		}
		case 34: {
			Get();
			break;
		}
		case 35: {
			Get();
			break;
		}
		case 36: {
			Get();
			break;
		}
		case 37: {
			Get();
			break;
		}
		case 38: {
			Get();
			break;
		}
		case 39: {
			Get();
			break;
		}
		case 40: {
			Get();
			break;
		}
		case 41: {
			Get();
			break;
		}
		case 42: {
			Get();
			break;
		}
		case 43: {
			Get();
			break;
		}
		case 44: {
			Get();
			break;
		}
		case 45: {
			Get();
			break;
		}
		case 46: {
			Get();
			break;
		}
		case 47: {
			Get();
			break;
		}
		case 48: {
			Get();
			break;
		}
		case 49: {
			Get();
			break;
		}
		case 50: {
			Get();
			break;
		}
		case 51: {
			Get();
			break;
		}
		case 52: {
			Get();
			break;
		}
		case 53: {
			Get();
			break;
		}
		case 54: {
			Get();
			break;
		}
		case 55: {
			Get();
			break;
		}
		case 56: {
			Get();
			break;
		}
		case 57: {
			Get();
			break;
		}
		case 58: {
			Get();
			break;
		}
		case 59: {
			Get();
			break;
		}
		case 60: {
			Get();
			break;
		}
		case 61: {
			Get();
			break;
		}
		case 62: {
			Get();
			break;
		}
		case 63: {
			Get();
			break;
		}
		case 64: {
			Get();
			break;
		}
		case 65: {
			Get();
			break;
		}
		case 66: {
			Get();
			break;
		}
		case 67: {
			Get();
			break;
		}
		case 68: {
			Get();
			break;
		}
		case 69: {
			Get();
			break;
		}
		case 70: {
			Get();
			break;
		}
		case 71: {
			Get();
			break;
		}
		case 72: {
			Get();
			break;
		}
		case 73: {
			Get();
			break;
		}
		case 74: {
			Get();
			break;
		}
		case 75: {
			Get();
			break;
		}
		case 76: {
			Get();
			break;
		}
		case 77: {
			Get();
			break;
		}
		case 78: {
			Get();
			break;
		}
		case 79: {
			Get();
			break;
		}
		case 80: {
			Get();
			break;
		}
		case 81: {
			Get();
			break;
		}
		case 82: {
			Get();
			break;
		}
		default: SynErr(167); break;
		}
	}

	void AttributeArguments() {
		bool nameFound = false; Hashtable htTmp = new Hashtable(); 
		Expect(92);
		if (StartOf(21)) {
			if (IsAssignment()) {
				nameFound = true; 
				Expect(1);
				Expect(96);
			}
			Expression(ref htTmp);
			while (la.kind == 86) {
				Get();
				if (IsAssignment()) {
					nameFound = true; 
					Expect(1);
					Expect(96);
				} else if (StartOf(21)) {
					if (nameFound) Error("no positional argument after named arguments"); 
				} else SynErr(168);
				Expression(ref htTmp);
			}
		}
		Expect(93);
	}

	void PrimitiveType(out String name) {
		name = ""; 
		if (StartOf(22)) {
			IntegralType(out name);
		} else if (la.kind == 32) {
			Get();
			name = "float"; 
		} else if (la.kind == 23) {
			Get();
			name = "double"; 
		} else if (la.kind == 19) {
			Get();
			name = "decimal"; 
		} else if (la.kind == 9) {
			Get();
			name = "bool"; 
		} else SynErr(169);
	}

	void PointerOrArray(ref TypeKind type) {
		while (IsPointerOrDims()) {
			if (la.kind == 116) {
				Get();
				type = TypeKind.pointer; 
			} else if (la.kind == 90) {
				Get();
				while (la.kind == 86) {
					Get();
				}
				Expect(91);
				type = TypeKind.array; 
			} else SynErr(170);
		}
	}

	void NonGenericClassType(out String name) {
		name = ""; 
		if (la.kind == 1) {
			NonGenericTypeName(out name);
		} else if (la.kind == 48 || la.kind == 65) {
			InternalClassType(out name);
		} else SynErr(171);
	}

	void InternalClassType(out String name) {
		name = ""; 
		if (la.kind == 48) {
			Get();
			name = "objet"; 
		} else if (la.kind == 65) {
			Get();
			name = "string"; 
		} else SynErr(172);
	}

	void TypeArgumentList(out String name) {
		TypeKind dummy;
		StringBuilder sb = new StringBuilder();
		
		Expect(122);
		Type(out dummy, false, out name);
		sb.Append("<").Append(name); 
		while (la.kind == 86) {
			Get();
			Type(out dummy, false, out name);
			sb.Append(",").Append(name); 
		}
		Expect(120);
		sb.Append(">");
		name = sb.ToString();
		
	}

	void Statement() {
		TypeKind dummy; String tname;
		Hashtable htTmp = new Hashtable();
		MethodInfo mInfo = ciCur.GetType() == typeof(MethodInfo) ? (MethodInfo)ciCur : null;
		
		if (la.kind == _ident && Peek(1).kind == _colon) {
			Expect(1);
			Expect(84);
			Statement();
		} else if (la.kind == 17) {
			Get();
			Type(out dummy, false, out tname);
			Expect(1);
			VariableInfo vInfo = new VariableInfo(t.val, tname, ciCur);
			ciCur.addVariable(vInfo);
			ciCur.statInfo.iStatements++;
			
			Expect(96);
			Expression(ref htTmp);
			if (mInfo!=null) mInfo.fSC += rechSC_Assign*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth); 
			while (la.kind == 86) {
				Get();
				Expect(1);
				vInfo = new VariableInfo(t.val, tname, ciCur);
				ciCur.addVariable(vInfo);
				
				Expect(96);
				Expression(ref htTmp);
				if (mInfo!=null) mInfo.fSC += rechSC_Assign*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth); 
			}
			Expect(87);
		} else if (IsLocalVarDecl()) {
			LocalVariableDeclaration();
			Expect(87);
			ciCur.statInfo.iStatements++; 
		} else if (StartOf(23)) {
			EmbeddedStatement();
		} else SynErr(173);
	}

	void EmbeddedStatement() {
		TypeKind type; String tname;
		Hashtable htTmp = new Hashtable();
		MethodInfo mInfo = (MethodInfo)ciCur;
		ciCur.statInfo.iStatements++;
		
		if (la.kind == 88) {
			Block();
		} else if (la.kind == 87) {
			Get();
		} else if (la.kind == _checked && Peek(1).kind == _lbrace) {
			Expect(15);
			mInfo.iCurDepth++;	
			Block();
			mInfo.iCurDepth--;	
		} else if (la.kind == _unchecked && Peek(1).kind == _lbrace) {
			Expect(75);
			mInfo.iCurDepth++;	
			Block();
			mInfo.iCurDepth--;	
		} else if (StartOf(21)) {
			StatementExpression();
			Expect(87);
		} else if (la.kind == 36) {
			Get();
			mInfo.fSC += rechSC_If*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth); 
			Expect(92);
			Expression(ref htTmp);
			Expect(93);
			mInfo.iCurDepth++;	
			EmbeddedStatement();
			mInfo.iCurDepth--;	
			if (la.kind == 24) {
				Get();
				mInfo.fSC += rechSC_Else*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth);
				mInfo.iCurDepth++;	
				EmbeddedStatement();
				mInfo.iCurDepth--;	
			}
		} else if (la.kind == 67) {
			Get();
			mInfo.fSC += rechSC_Select*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth); 
			Expect(92);
			Expression(ref htTmp);
			Expect(93);
			Expect(88);
			while (la.kind == 12 || la.kind == 20) {
				mInfo.iCurDepth++;	
				SwitchSection();
				mInfo.iCurDepth--;	
			}
			Expect(89);
		} else if (la.kind == 82) {
			Get();
			mInfo.fSC += rechSC_While*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth); 
			Expect(92);
			Expression(ref htTmp);
			Expect(93);
			mInfo.iCurDepth++;	
			EmbeddedStatement();
			mInfo.iCurDepth--;	
		} else if (la.kind == 22) {
			Get();
			mInfo.iCurDepth++;	
			EmbeddedStatement();
			Expect(82);
			mInfo.iCurDepth--;
			mInfo.fSC += rechSC_While*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth); 
			Expect(92);
			Expression(ref htTmp);
			Expect(93);
			Expect(87);
		} else if (la.kind == 33) {
			Get();
			mInfo.fSC += rechSC_For*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth); 
			Expect(92);
			if (StartOf(24)) {
				ForInitializer();
			}
			Expect(87);
			if (StartOf(21)) {
				Expression(ref htTmp);
			}
			Expect(87);
			if (StartOf(21)) {
				ForIterator();
			}
			Expect(93);
			mInfo.iCurDepth++;	
			EmbeddedStatement();
			mInfo.iCurDepth--;	
		} else if (la.kind == 34) {
			Get();
			mInfo.fSC += rechSC_For*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth); 
			Expect(92);
			Type(out type, false, out tname);
			Expect(1);
			VariableInfo vInfo = new VariableInfo(t.val, tname, ciCur);
			ciCur.addVariable(vInfo);
			
			Expect(38);
			Expression(ref htTmp);
			Expect(93);
			mInfo.iCurDepth++;	
			EmbeddedStatement();
			mInfo.iCurDepth--;	
		} else if (la.kind == 10) {
			Get();
			mInfo.fSC += rechSC_Break*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth); 
			Expect(87);
		} else if (la.kind == 18) {
			Get();
			mInfo.fSC += rechSC_Continue*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth); 
			Expect(87);
		} else if (la.kind == 35) {
			Get();
			mInfo.fSC += rechSC_Goto*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth); 
			if (la.kind == 1) {
				Get();
			} else if (la.kind == 12) {
				Get();
				Expression(ref htTmp);
			} else if (la.kind == 20) {
				Get();
			} else SynErr(174);
			Expect(87);
		} else if (la.kind == 58) {
			Get();
			mInfo.fSC += rechSC_Return*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth); 
			if (StartOf(21)) {
				htTmp = new Hashtable(); 
				Expression(ref htTmp);
				((MethodInfo) ciCur).addVarAcc(htTmp, true);
				htTmp = new Hashtable(); htTmp["return"] = 1;
				((MethodInfo) ciCur).addVarAcc(htTmp, false); 
			}
			Expect(87);
		} else if (la.kind == 69) {
			Get();
			mInfo.fSC += rechSC_Throw*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth); 
			if (StartOf(21)) {
				Expression(ref htTmp);
			}
			Expect(87);
		} else if (la.kind == 71) {
			Get();
			mInfo.fSC += rechSC_Try*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth);
			mInfo.iCurDepth++;	
			Block();
			mInfo.iCurDepth--;	
			if (la.kind == 13) {
				CatchClauses();
				if (la.kind == 30) {
					Get();
					mInfo.iCurDepth++;	
					Block();
					mInfo.iCurDepth--;	
				}
			} else if (la.kind == 30) {
				Get();
				mInfo.iCurDepth++;	
				Block();
				mInfo.iCurDepth--;	
			} else SynErr(175);
		} else if (la.kind == 43) {
			Get();
			Expect(92);
			Expression(ref htTmp);
			Expect(93);
			mInfo.iCurDepth++;	
			EmbeddedStatement();
			mInfo.iCurDepth--;	
		} else if (la.kind == 78) {
			Get();
			Expect(92);
			ResourceAcquisition();
			Expect(93);
			mInfo.iCurDepth++;	
			EmbeddedStatement();
			mInfo.iCurDepth--;	
		} else if (la.kind == 127) {
			Get();
			if (la.kind == 58) {
				Get();
				Expression(ref htTmp);
			} else if (la.kind == 10) {
				Get();
			} else SynErr(176);
			Expect(87);
		} else if (la.kind == 76) {
			Get();
			mInfo.iCurDepth++;	
			Block();
			mInfo.iCurDepth--;	
		} else if (la.kind == 31) {
			Get();
			Expect(92);
			Type(out type, false, out tname);
			if (type != TypeKind.pointer) Error("can only fix pointer types"); 
			Expect(1);
			VariableInfo vInfo = new VariableInfo(t.val, tname, ciCur);
			ciCur.addVariable(vInfo);
			
			Expect(96);
			mInfo.fSC += rechSC_Assign*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth); 
			Expression(ref htTmp);
			while (la.kind == 86) {
				Get();
				Expect(1);
				vInfo = new VariableInfo(t.val, tname, ciCur);
				ciCur.addVariable(vInfo);
				
				Expect(96);
				Expression(ref htTmp);
			}
			Expect(93);
			mInfo.iCurDepth++;	
			EmbeddedStatement();
			mInfo.iCurDepth--;	
		} else SynErr(177);
	}

	void StatementExpression() {
		bool isAssignment = assnStartOp[la.kind] || IsTypeCast();
		Hashtable htTmp = new Hashtable();
		MethodInfo mInfo = (MethodInfo) ciCur;
		
		Unary(ref htTmp);
		if (StartOf(25)) {
			AssignmentOperator();
			if (t.val != "=") //keine reine Zuweisung, kein +=, -= etc.
			mInfo.addVarAcc(htTmp, true);
			mInfo.addVarAcc(htTmp, false);
			htTmp = new Hashtable();
			mInfo.fSC += rechSC_Assign*(float)Math.Pow(rechSC_Base, mInfo.iCurDepth);
			
			Expression(ref htTmp);
		} else if (la.kind == 86 || la.kind == 87 || la.kind == 93) {
			if (isAssignment) Error("error in assignment."); 
		} else SynErr(178);
		mInfo.addVarAcc(htTmp, true); 
	}

	void SwitchSection() {
		SwitchLabel();
		while (la.kind == _case || (la.kind == _default && Peek(1).kind == _colon)) {
			SwitchLabel();
		}
		Statement();
		while (IsNoSwitchLabelOrRBrace()) {
			Statement();
		}
	}

	void ForInitializer() {
		if (IsLocalVarDecl()) {
			LocalVariableDeclaration();
		} else if (StartOf(21)) {
			StatementExpression();
			while (la.kind == 86) {
				Get();
				StatementExpression();
			}
		} else SynErr(179);
	}

	void ForIterator() {
		StatementExpression();
		while (la.kind == 86) {
			Get();
			StatementExpression();
		}
	}

	void CatchClauses() {
		String tname;
		MethodInfo mInfo = (MethodInfo) ciCur;
		
		Expect(13);
		if (la.kind == 88) {
			mInfo.iCurDepth++;	
			Block();
			mInfo.iCurDepth--;	
		} else if (la.kind == 92) {
			Get();
			ClassType(out tname);
			if (la.kind == 1) {
				Get();
				VariableInfo vInfo = new VariableInfo(t.val, tname, ciCur);
				ciCur.addVariable(vInfo);
				
			}
			Expect(93);
			mInfo.iCurDepth++;	
			Block();
			mInfo.iCurDepth--;	
			if (la.kind == 13) {
				CatchClauses();
			}
		} else SynErr(180);
	}

	void ResourceAcquisition() {
		Hashtable htTmp = new Hashtable(); 
		if (IsLocalVarDecl()) {
			LocalVariableDeclaration();
		} else if (StartOf(21)) {
			Expression(ref htTmp);
		} else SynErr(181);
	}

	void Unary(ref Hashtable htVarAcc) {
		TypeKind dummy; String tname;
		MethodInfo mInfo = ciCur.GetType() == typeof(MethodInfo) ? (MethodInfo) ciCur : null;
		
		while (unaryHead[la.kind] || IsTypeCast()) {
			switch (la.kind) {
			case 108: {
				Get();
				break;
			}
			case 111: {
				Get();
				if (mInfo!=null) mInfo.fEC += rechEC_Add*(float)Math.Pow(rechEC_Base, mInfo.iExpDepth); 
				break;
			}
			case 99: {
				Get();
				if (mInfo!=null) mInfo.fEC += rechEC_Not*(float)Math.Pow(rechEC_Base, mInfo.iExpDepth); 
				break;
			}
			case 94: {
				Get();
				if (mInfo!=null) mInfo.fEC += rechSC_Bin*(float)Math.Pow(rechEC_Base, mInfo.iExpDepth); 
				break;
			}
			case 110: {
				Get();
				if (mInfo!=null) mInfo.fEC += rechEC_Add*(float)Math.Pow(rechEC_Base, mInfo.iExpDepth); 
				break;
			}
			case 113: {
				Get();
				if (mInfo!=null) mInfo.fEC += rechEC_Add*(float)Math.Pow(rechEC_Base, mInfo.iExpDepth); 
				break;
			}
			case 116: {
				Get();
				if (mInfo!=null) mInfo.fEC += rechEC_Deref*(float)Math.Pow(rechEC_Base, mInfo.iExpDepth); 
				break;
			}
			case 100: {
				Get();
				if (mInfo!=null) mInfo.fEC += rechEC_Deref*(float)Math.Pow(rechEC_Base, mInfo.iExpDepth); 
				break;
			}
			case 92: {
				Get();
				NonGenericType(out dummy, false, out tname);
				Expect(93);
				break;
			}
			default: SynErr(182); break;
			}
		}
		Primary(ref htVarAcc);
	}

	void AssignmentOperator() {
		MethodInfo mInfo = (MethodInfo)ciCur;
		float fTmp = (float)Math.Pow(rechEC_Base, mInfo.iExpDepth);
		
		switch (la.kind) {
		case 96: {
			Get();
			break;
		}
		case 109: {
			Get();
			mInfo.fEC += rechEC_Add*fTmp; 
			break;
		}
		case 112: {
			Get();
			mInfo.fEC += rechEC_Add*fTmp; 
			break;
		}
		case 117: {
			Get();
			mInfo.fEC += rechEC_Mul*fTmp; 
			break;
		}
		case 115: {
			Get();
			mInfo.fEC += rechEC_Mul*fTmp; 
			break;
		}
		case 119: {
			Get();
			mInfo.fEC += rechEC_Mod*fTmp; 
			break;
		}
		case 101: {
			Get();
			mInfo.fEC += rechSC_Bin*fTmp; 
			break;
		}
		case 104: {
			Get();
			mInfo.fEC += rechSC_Bin*fTmp; 
			break;
		}
		case 107: {
			Get();
			mInfo.fEC += rechSC_Bin*fTmp; 
			break;
		}
		case 125: {
			Get();
			mInfo.fEC += rechEC_Shft*fTmp; 
			break;
		}
		case 120: {
			Get();
			int pos = t.pos; 
			Expect(121);
			if (pos+1 < t.pos) Error("no whitespace allowed in right shift assignment");
			// Es wurde ein Right Shift Assign gefunden -> Statement-Info's anpassen!
			int i = (int) ciCur.statInfo.htOperators[">"];
			ciCur.statInfo.htOperators[">"] = i-1;
			i = (int) ciCur.statInfo.htOperators[">="];
			ciCur.statInfo.htOperators[">="] = i-1;
			if (ciCur.statInfo.htOperators.ContainsKey(">>=")) {
				i = (int) ciCur.statInfo.htOperators[">>="];
				ciCur.statInfo.htOperators[">>="] = i+1;
			}
			else ciCur.statInfo.htOperators[">>="] = 1;
			mInfo.fEC += rechEC_Shft*fTmp;
			
			break;
		}
		default: SynErr(183); break;
		}
	}

	void SwitchLabel() {
		Hashtable htTmp = new Hashtable(); 
		if (la.kind == 12) {
			Get();
			ciCur.statInfo.iCaseInSwitch++; 
			Expression(ref htTmp);
			Expect(84);
		} else if (la.kind == 20) {
			Get();
			Expect(84);
		} else SynErr(184);
	}

	void OrExpr(ref Hashtable htVarAcc) {
		AndExpr(ref htVarAcc);
		while (la.kind == 105) {
			Get();
			if (ciCur.GetType() == typeof(MethodInfo))
			((MethodInfo)ciCur).fEC += rechEC_And*(float)Math.Pow(rechEC_Base, ((MethodInfo)ciCur).iExpDepth);
			
			Unary(ref htVarAcc);
			AndExpr(ref htVarAcc);
		}
	}

	void AndExpr(ref Hashtable htVarAcc) {
		BitOrExpr(ref htVarAcc);
		while (la.kind == 102) {
			Get();
			if (ciCur.GetType() == typeof(MethodInfo))
			((MethodInfo)ciCur).fEC += rechEC_And*(float)Math.Pow(rechEC_Base, ((MethodInfo)ciCur).iExpDepth);
			
			Unary(ref htVarAcc);
			BitOrExpr(ref htVarAcc);
		}
	}

	void BitOrExpr(ref Hashtable htVarAcc) {
		BitXorExpr(ref htVarAcc);
		while (la.kind == 103) {
			Get();
			if (ciCur.GetType() == typeof(MethodInfo))
			((MethodInfo)ciCur).fEC += rechSC_Bin*(float)Math.Pow(rechEC_Base, ((MethodInfo)ciCur).iExpDepth);
			
			Unary(ref htVarAcc);
			BitXorExpr(ref htVarAcc);
		}
	}

	void BitXorExpr(ref Hashtable htVarAcc) {
		BitAndExpr(ref htVarAcc);
		while (la.kind == 106) {
			Get();
			if (ciCur.GetType() == typeof(MethodInfo))
			((MethodInfo)ciCur).fEC += rechSC_Bin*(float)Math.Pow(rechEC_Base, ((MethodInfo)ciCur).iExpDepth);
			
			Unary(ref htVarAcc);
			BitAndExpr(ref htVarAcc);
		}
	}

	void BitAndExpr(ref Hashtable htVarAcc) {
		EqlExpr(ref htVarAcc);
		while (la.kind == 100) {
			Get();
			if (ciCur.GetType() == typeof(MethodInfo))
			((MethodInfo)ciCur).fEC += rechSC_Bin*(float)Math.Pow(rechEC_Base, ((MethodInfo)ciCur).iExpDepth);
			
			Unary(ref htVarAcc);
			EqlExpr(ref htVarAcc);
		}
	}

	void EqlExpr(ref Hashtable htVarAcc) {
		RelExpr(ref htVarAcc);
		while (la.kind == 97 || la.kind == 98) {
			if (la.kind == 98) {
				Get();
			} else {
				Get();
			}
			if (ciCur.GetType() == typeof(MethodInfo))
			((MethodInfo)ciCur).fEC += rechEC_Rel*(float)Math.Pow(rechEC_Base, ((MethodInfo)ciCur).iExpDepth);
			
			Unary(ref htVarAcc);
			RelExpr(ref htVarAcc);
		}
	}

	void RelExpr(ref Hashtable htVarAcc) {
		TypeKind dummy; String tname; 
		ShiftExpr(ref htVarAcc);
		while (StartOf(26)) {
			if (StartOf(27)) {
				if (la.kind == 122) {
					Get();
				} else if (la.kind == 120) {
					Get();
				} else if (la.kind == 123) {
					Get();
				} else if (la.kind == 121) {
					Get();
				} else SynErr(185);
				if (ciCur.GetType() == typeof(MethodInfo))
				((MethodInfo)ciCur).fEC += rechEC_Rel*(float)Math.Pow(rechEC_Base, ((MethodInfo)ciCur).iExpDepth);
				
				Unary(ref htVarAcc);
				ShiftExpr(ref htVarAcc);
			} else {
				if (la.kind == 42) {
					Get();
				} else if (la.kind == 7) {
					Get();
				} else SynErr(186);
				NonGenericType(out dummy, false, out tname);
			}
		}
	}

	void ShiftExpr(ref Hashtable htVarAcc) {
		AddExpr(ref htVarAcc);
		while (IsShift()) {
			if (la.kind == 124) {
				Get();
			} else if (la.kind == 120) {
				Get();
				Expect(120);
				if (ciCur.GetType() == typeof(MethodInfo))
				((MethodInfo)ciCur).fEC += rechEC_Shft*(float)Math.Pow(rechEC_Base, ((MethodInfo)ciCur).iExpDepth);
				int i = (int) ciCur.statInfo.htOperators[">"];
				ciCur.statInfo.htOperators[">"] = i - 2;
				if (ciCur.statInfo.htOperators.ContainsKey(">>")) {
					i = (int) ciCur.statInfo.htOperators[">>"];
					ciCur.statInfo.htOperators[">>"] = i + 1;
				}
				else ciCur.statInfo.htOperators[">>"] = 1;
				//Halstead Info's anpassen
				ciCur.hInfo.replaceRShift();
				
			} else SynErr(187);
			Unary(ref htVarAcc);
			AddExpr(ref htVarAcc);
		}
	}

	void AddExpr(ref Hashtable htVarAcc) {
		MulExpr(ref htVarAcc);
		while (la.kind == 108 || la.kind == 111) {
			if (la.kind == 108) {
				Get();
			} else {
				Get();
			}
			if (ciCur.GetType() == typeof(MethodInfo))
			((MethodInfo)ciCur).fEC += rechEC_Add*(float)Math.Pow(rechEC_Base, ((MethodInfo)ciCur).iExpDepth);
			
			Unary(ref htVarAcc);
			MulExpr(ref htVarAcc);
		}
	}

	void MulExpr(ref Hashtable htVarAcc) {
		MethodInfo mInfo = ciCur.GetType() == typeof(MethodInfo) ? (MethodInfo)ciCur : null; 
		while (la.kind == 114 || la.kind == 116 || la.kind == 118) {
			if (la.kind == 116) {
				Get();
				if (mInfo!=null) mInfo.fEC += rechEC_Mul*(float)Math.Pow(rechEC_Base, mInfo.iExpDepth); 
			} else if (la.kind == 114) {
				Get();
				if (mInfo!=null) mInfo.fEC += rechEC_Mul*(float)Math.Pow(rechEC_Base, mInfo.iExpDepth); 
			} else {
				Get();
				if (mInfo!=null) mInfo.fEC += rechEC_Mod*(float)Math.Pow(rechEC_Base, mInfo.iExpDepth); 
			}
			Unary(ref htVarAcc);
		}
	}

	void Primary(ref Hashtable htVarAcc) {
		TypeKind type; String tname;
		bool isArrayCreation = false;
		StringBuilder sbMethod = new StringBuilder();
		int ArgCount = 0;
		MethodInfo miMeth = ciCur.GetType() == typeof(MethodInfo) ? (MethodInfo) ciCur : null;
		
		switch (la.kind) {
		case 2: case 3: case 4: case 5: case 29: case 47: case 70: {
			Literal();
			break;
		}
		case 92: {
			Get();
			Expression(ref htVarAcc);
			Expect(93);
			break;
		}
		case 9: case 11: case 14: case 19: case 23: case 32: case 39: case 44: case 48: case 59: case 61: case 65: case 73: case 74: case 77: {
			switch (la.kind) {
			case 9: {
				Get();
				break;
			}
			case 11: {
				Get();
				sbMethod.Append("byte"); 
				break;
			}
			case 14: {
				Get();
				sbMethod.Append("char"); 
				break;
			}
			case 19: {
				Get();
				sbMethod.Append("decimal"); 
				break;
			}
			case 23: {
				Get();
				sbMethod.Append("double"); 
				break;
			}
			case 32: {
				Get();
				sbMethod.Append("float"); 
				break;
			}
			case 39: {
				Get();
				sbMethod.Append("int"); 
				break;
			}
			case 44: {
				Get();
				sbMethod.Append("long"); 
				break;
			}
			case 48: {
				Get();
				sbMethod.Append("object"); 
				break;
			}
			case 59: {
				Get();
				sbMethod.Append("sbyte"); 
				break;
			}
			case 61: {
				Get();
				sbMethod.Append("short"); 
				break;
			}
			case 65: {
				Get();
				sbMethod.Append("string"); 
				break;
			}
			case 73: {
				Get();
				sbMethod.Append("uint"); 
				break;
			}
			case 74: {
				Get();
				sbMethod.Append("ulong"); 
				break;
			}
			case 77: {
				Get();
				sbMethod.Append("ushort"); 
				break;
			}
			}
			Expect(83);
			Expect(1);
			if (miMeth!=null) miMeth.fEC += rechEC_Fld*(float)Math.Pow(rechEC_Base, miMeth.iExpDepth);
			sbMethod.Append(".").Append(t.val);
			
			break;
		}
		case 1: {
			Get();
			sbMethod.Append(t.val); 
			if (la.kind == 85) {
				Get();
				Expect(1);
				sbMethod.Append(".").Append(t.val); 
				Expect(83);
				Expect(1);
				if (miMeth!=null) miMeth.fEC += rechEC_Fld*(float)Math.Pow(rechEC_Base, miMeth.iExpDepth);
				sbMethod.Append(".").Append(t.val); 
			}
			break;
		}
		case 68: {
			Get();
			sbMethod.Append("this"); 
			break;
		}
		case 8: {
			Get();
			sbMethod.Append("base"); 
			if (la.kind == 83) {
				Get();
				Expect(1);
				if (miMeth!=null) miMeth.fEC += rechEC_Fld*(float)Math.Pow(rechEC_Base, miMeth.iExpDepth);
				sbMethod.Append(".").Append(t.val);
				
			} else if (la.kind == 90) {
				Get();
				Expression(ref htVarAcc);
				while (la.kind == 86) {
					Get();
					Expression(ref htVarAcc);
				}
				Expect(91);
			} else SynErr(188);
			break;
		}
		case 46: {
			Get();
			Type(out type, false, out tname);
			sbMethod.Append(tname); 
			if (la.kind == 92) {
				Get();
				if (StartOf(15)) {
					Argument();
					ArgCount = 1; 
					while (la.kind == 86) {
						Get();
						Argument();
						ArgCount++; 
					}
				}
				Expect(93);
				sbMethod.AppendFormat("({0})", ArgCount); 
				if (miMeth!=null) {
					String sMethod = sbMethod.ToString();
					miMeth.fSC += (rechSC_Method+ArgCount)*(float)Math.Pow(rechSC_Base, miMeth.iCurDepth);
					miMeth.fEC += (rechEC_Method+ArgCount)*(float)Math.Pow(rechEC_Base, miMeth.iExpDepth);
					if (miMeth.htMethodCalls.ContainsKey(sMethod)) {
						int i = (int) miMeth.htMethodCalls[sMethod];
						miMeth.htMethodCalls[sMethod] = i + 1;
					}
					else {
						miMeth.htMethodCalls[sMethod] = 1;
					}
				}
				
			} else if (la.kind == 90) {
				Get();
				Expression(ref htVarAcc);
				while (la.kind == 86) {
					Get();
					Expression(ref htVarAcc);
				}
				Expect(91);
				while (IsDims()) {
					Expect(90);
					while (la.kind == 86) {
						Get();
					}
					Expect(91);
				}
				if (la.kind == 88) {
					ArrayInitializer();
				}
				isArrayCreation = true; 
			} else if (la.kind == 88) {
				ArrayInitializer();
				if (type != TypeKind.array) Error("array type expected");
				isArrayCreation = true;
				
			} else SynErr(189);
			break;
		}
		case 72: {
			Get();
			Expect(92);
			Type(out type, true, out tname);
			Expect(93);
			break;
		}
		case 15: {
			Get();
			Expect(92);
			Expression(ref htVarAcc);
			Expect(93);
			break;
		}
		case 75: {
			Get();
			Expect(92);
			Expression(ref htVarAcc);
			Expect(93);
			break;
		}
		case 20: {
			Get();
			Expect(92);
			Primary(ref htVarAcc);
			Expect(93);
			break;
		}
		case 21: {
			Get();
			if (la.kind == 92) {
				Get();
				if (StartOf(13)) {
					AnonymousMethodParameter();
					while (la.kind == 86) {
						Get();
						AnonymousMethodParameter();
					}
				}
				Expect(93);
			}
			Block();
			break;
		}
		case 62: {
			Get();
			Expect(92);
			Type(out type, false, out tname);
			Expect(93);
			break;
		}
		default: SynErr(190); break;
		}
		while (StartOf(28)) {
			switch (la.kind) {
			case 110: {
				Get();
				if (miMeth!=null) miMeth.fEC += rechEC_Add*(float)Math.Pow(rechEC_Base, miMeth.iExpDepth);
				if(miMeth != null && sbMethod.Length != 0 && sbMethod.ToString().IndexOf(")") == -1) {
					Hashtable htTmp = new Hashtable(); htTmp[sbMethod.ToString()] = 1;
					miMeth.addVarAcc(htTmp, false); miMeth.addVarAcc(htTmp, true);
				}
				
				break;
			}
			case 113: {
				Get();
				if (miMeth!=null) miMeth.fEC += rechEC_Add*(float)Math.Pow(rechEC_Base, miMeth.iExpDepth);
				if(miMeth != null && sbMethod.Length != 0 && sbMethod.ToString().IndexOf(")") == -1) {
					Hashtable htTmp = new Hashtable(); htTmp[sbMethod.ToString()] = 1;
					miMeth.addVarAcc(htTmp, false); miMeth.addVarAcc(htTmp, true);
				}
				
				break;
			}
			case 128: {
				Get();
				Expect(1);
				if (miMeth!=null) miMeth.fEC += rechEC_Fld*(float)Math.Pow(rechEC_Base, miMeth.iExpDepth);
				sbMethod.Append(".").Append(t.val);
				
				break;
			}
			case 83: {
				Get();
				Expect(1);
				if (miMeth!=null) miMeth.fEC += rechEC_Fld*(float)Math.Pow(rechEC_Base, miMeth.iExpDepth);
				sbMethod.Append(".").Append(t.val);
				
				break;
			}
			case 92: {
				Get();
				if (StartOf(15)) {
					Argument();
					ArgCount = 1; 
					while (la.kind == 86) {
						Get();
						Argument();
						ArgCount++; 
					}
				}
				Expect(93);
				sbMethod.AppendFormat("({0})", ArgCount);
				String sMethod = sbMethod.ToString();
				if (miMeth != null) {
					miMeth.fSC += (rechSC_Method+ArgCount)*(float)Math.Pow(rechSC_Base, miMeth.iCurDepth);
					miMeth.fEC += (rechEC_Method+ArgCount)*(float)Math.Pow(rechEC_Base, miMeth.iExpDepth);
					if (miMeth.htMethodCalls.ContainsKey(sMethod)) {
						int i = (int) miMeth.htMethodCalls[sMethod];
						miMeth.htMethodCalls[sMethod] = i + 1;
					}
					else {
						miMeth.htMethodCalls[sMethod] = 1;
					}
				}
				sbMethod = new StringBuilder();
				
				break;
			}
			case 90: {
				if (isArrayCreation) Error("element access not allow on array creation"); 
				Get();
				Expression(ref htVarAcc);
				if (miMeth!=null) miMeth.fEC += rechEC_Indx*(float)Math.Pow(rechEC_Base, miMeth.iExpDepth); 
				while (la.kind == 86) {
					Get();
					Expression(ref htVarAcc);
					if (miMeth!=null) miMeth.fEC += rechEC_Indx*(float)Math.Pow(rechEC_Base, miMeth.iExpDepth); 
				}
				Expect(91);
				break;
			}
			}
		}
		if(miMeth != null &&
		sbMethod.Length != 0 &&
		sbMethod.ToString().IndexOf(")") == -1) {
		htVarAcc[sbMethod.ToString()] = 1;
		if (!miMeth.htOperands.ContainsKey(sbMethod.ToString())) {
			miMeth.htOperands[sbMethod.ToString()] = 1;
		}
		}
		
	}

	void Literal() {
		switch (la.kind) {
		case 2: {
			Get();
			break;
		}
		case 3: {
			Get();
			break;
		}
		case 4: {
			Get();
			break;
		}
		case 5: {
			Get();
			break;
		}
		case 70: {
			Get();
			break;
		}
		case 29: {
			Get();
			break;
		}
		case 47: {
			Get();
			break;
		}
		default: SynErr(191); break;
		}
	}

	void AnonymousMethodParameter() {
		TypeKind dummy; String tname; 
		if (la.kind == 50 || la.kind == 57) {
			if (la.kind == 57) {
				Get();
			} else {
				Get();
			}
		}
		Type(out dummy, false, out tname);
		Expect(1);
	}



	public void Parse() {
		la = new Token();
		la.val = "";		
		Get();
		CS2();

    Expect(0);
	}
	
	bool[,] set = {
		{T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x},
		{x,x,x,x, x,x,T,x, x,x,x,x, x,x,x,x, T,x,x,x, x,T,x,x, x,T,x,x, T,x,x,x, x,x,x,x, x,x,x,x, T,T,x,x, x,T,T,x, x,x,x,T, x,T,T,T, T,x,x,x, T,x,x,x, T,x,T,x, x,x,x,x, x,x,x,x, T,x,x,T, x,T,x,x, x,x,x,x, x,x,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,T,x, x,x,x},
		{x,x,x,x, x,x,T,x, x,x,x,x, x,x,x,x, T,x,x,x, x,T,x,x, x,T,x,x, T,x,x,x, x,x,x,x, x,x,x,x, T,T,x,x, x,x,T,x, x,x,x,T, x,T,T,T, T,x,x,x, T,x,x,x, T,x,T,x, x,x,x,x, x,x,x,x, T,x,x,T, x,T,x,x, x,x,x,x, x,x,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,T,x, x,x,x},
		{x,x,x,x, x,x,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,T, T,T,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x},
		{x,x,x,x, x,x,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, T,x,x,x, x,x,x,x, x,x,x,x, x,T,x,x, x,x,T,x, x,x,x,T, x,T,T,T, T,x,x,x, T,x,x,x, T,x,x,x, x,x,x,x, x,x,x,x, T,x,x,T, x,T,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x},
		{x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,T,x, x,x,x},
		{x,T,x,x, x,x,x,x, x,T,x,T, x,x,T,x, x,x,x,T, x,x,x,T, x,x,T,x, x,x,x,x, T,x,x,x, x,x,x,T, x,x,x,x, T,x,T,x, T,x,x,x, x,x,x,x, x,x,x,T, x,T,x,x, x,T,x,x, x,x,x,x, x,T,T,x, x,T,x,x, T,x,x,x, x,x,x,x, x,x,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x},
		{x,T,x,x, x,x,x,x, x,T,x,T, x,x,T,x, x,x,x,T, x,x,x,T, x,x,x,x, x,x,x,x, T,x,x,x, x,x,x,T, x,x,x,x, T,x,x,x, T,x,T,x, T,x,x,x, x,T,x,T, x,T,x,x, x,T,x,x, x,x,x,x, x,T,T,x, x,T,x,x, T,x,x,x, x,x,x,x, x,x,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x},
		{x,T,x,x, x,x,x,x, x,x,x,x, x,x,x,x, T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,T,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x},
		{x,T,x,x, x,x,T,x, x,T,x,T, x,x,T,x, T,T,x,T, x,T,x,T, x,T,T,T, T,x,x,x, T,x,x,x, x,T,x,T, T,T,x,x, T,x,T,x, T,x,x,T, x,T,T,T, T,x,x,T, T,T,x,x, T,T,T,x, x,x,x,x, x,T,T,x, T,T,x,T, T,T,x,x, x,x,x,x, x,x,T,x, x,x,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,T,x, x,x,x},
		{x,T,x,x, x,x,T,x, x,T,x,T, x,x,T,x, T,T,x,T, x,T,x,T, x,T,T,T, T,x,x,x, T,x,x,x, x,T,x,T, T,T,x,x, T,x,T,x, T,x,x,T, x,T,T,T, T,x,x,T, T,T,x,x, T,T,T,x, x,x,x,x, x,T,T,x, T,T,x,T, T,T,x,x, x,x,x,x, x,x,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,T,x, x,x,x},
		{x,T,x,x, x,x,x,x, x,T,x,T, x,x,T,x, x,x,x,T, x,x,x,T, x,x,x,x, x,x,x,x, T,x,x,x, x,x,x,T, x,x,x,x, T,x,x,x, T,x,x,x, x,x,x,x, x,x,x,T, x,T,x,x, x,T,x,x, x,x,x,x, x,T,T,x, x,T,x,x, T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x},
		{x,x,x,x, x,x,x,x, x,T,x,T, x,x,T,x, x,x,x,T, x,x,x,T, x,x,x,x, x,x,x,x, T,x,x,x, x,x,x,T, x,x,x,x, T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,T, x,T,x,x, x,x,x,x, x,x,x,x, x,T,T,x, x,T,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x},
		{x,T,x,x, x,x,x,x, x,T,x,T, x,x,T,x, x,x,x,T, x,x,x,T, x,x,x,x, x,x,x,x, T,x,x,x, x,x,x,T, x,x,x,x, T,x,x,x, T,x,T,x, x,x,x,x, x,T,x,T, x,T,x,x, x,T,x,x, x,x,x,x, x,T,T,x, x,T,x,x, T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x},
		{x,T,x,x, x,x,x,x, x,T,x,T, x,x,T,x, T,T,x,T, x,T,x,T, x,T,T,T, x,x,x,x, T,x,x,x, x,T,x,T, T,x,x,x, T,x,x,x, T,x,x,x, x,x,x,x, x,x,x,T, x,T,x,x, x,T,T,x, x,x,x,x, x,T,T,x, x,T,x,x, T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,T,x, x,x,x},
		{x,T,T,T, T,T,x,x, T,T,x,T, x,x,T,T, x,x,x,T, T,T,x,T, x,x,x,x, x,T,x,x, T,x,x,x, x,x,x,T, x,x,x,x, T,x,T,T, T,x,T,x, x,x,x,x, x,T,x,T, x,T,T,x, x,T,x,x, T,x,T,x, T,T,T,T, x,T,x,x, x,x,x,x, x,x,x,x, x,x,x,x, T,x,T,x, x,x,x,T, T,x,x,x, x,x,x,x, T,x,T,T, x,T,x,x, T,x,x,x, x,x,x,x, x,x,x,x, x,x,x},
		{x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, T,x,x,x, x,T,x,x, x,T,x,x, x,x,x,x, x,x,x,x, x,x,x,x, T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,T,x, x,x,x},
		{x,T,T,T, T,T,x,x, T,T,T,T, x,x,T,T, x,T,T,T, T,T,T,T, x,x,x,x, x,T,x,T, T,T,T,T, T,x,x,T, x,x,x,T, T,x,T,T, T,x,x,x, x,x,x,x, x,x,T,T, x,T,T,x, x,T,x,T, T,T,T,T, T,T,T,T, T,T,T,x, T,x,T,x, x,x,x,T, T,x,x,x, T,x,T,x, x,x,x,T, T,x,x,x, x,x,x,x, T,x,T,T, x,T,x,x, T,x,x,x, x,x,x,x, x,x,x,T, x,x,x},
		{x,x,x,x, x,x,x,T, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, T,x,T,T, x,T,x,T, x,T,x,T, x,T,T,x, T,x,T,T, x,T,T,x, T,x,x,T, x,x,T,x, T,x,T,x, T,T,T,T, T,x,x,x, x,x,x},
		{x,T,x,x, x,x,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, T,x,x,x, x,x,x,x, x,x,x,x, x,T,x,x, x,x,T,x, x,x,x,T, x,T,T,T, T,x,x,x, T,x,x,x, T,x,x,x, x,x,x,x, x,x,x,x, T,x,x,T, x,T,x,x, x,x,x,x, x,x,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x},
		{x,T,T,T, T,T,x,x, T,T,x,T, x,x,T,T, x,x,x,T, T,T,x,T, x,x,x,x, x,T,x,x, T,x,x,x, x,x,x,T, x,x,x,x, T,x,T,T, T,x,x,x, x,x,x,x, x,x,x,T, x,T,T,x, x,T,x,x, T,x,T,x, T,T,T,T, x,T,x,x, x,x,x,x, x,x,x,x, T,x,x,x, T,x,T,x, x,x,x,T, T,x,x,x, x,x,x,x, T,x,T,T, x,T,x,x, T,x,x,x, x,x,x,x, x,x,x,x, x,x,x},
		{x,T,T,T, T,T,x,x, T,T,x,T, x,x,T,T, x,x,x,T, T,T,x,T, x,x,x,x, x,T,x,x, T,x,x,x, x,x,x,T, x,x,x,x, T,x,T,T, T,x,x,x, x,x,x,x, x,x,x,T, x,T,T,x, x,T,x,x, T,x,T,x, T,T,T,T, x,T,x,x, x,x,x,x, x,x,x,x, x,x,x,x, T,x,T,x, x,x,x,T, T,x,x,x, x,x,x,x, T,x,T,T, x,T,x,x, T,x,x,x, x,x,x,x, x,x,x,x, x,x,x},
		{x,x,x,x, x,x,x,x, x,x,x,T, x,x,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,T, x,x,x,x, T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,T, x,T,x,x, x,x,x,x, x,x,x,x, x,T,T,x, x,T,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x},
		{x,T,T,T, T,T,x,x, T,T,T,T, x,x,T,T, x,x,T,T, T,T,T,T, x,x,x,x, x,T,x,T, T,T,T,T, T,x,x,T, x,x,x,T, T,x,T,T, T,x,x,x, x,x,x,x, x,x,T,T, x,T,T,x, x,T,x,T, T,T,T,T, T,T,T,T, T,T,T,x, x,x,T,x, x,x,x,T, T,x,x,x, T,x,T,x, x,x,x,T, T,x,x,x, x,x,x,x, T,x,T,T, x,T,x,x, T,x,x,x, x,x,x,x, x,x,x,T, x,x,x},
		{x,T,T,T, T,T,x,x, T,T,x,T, x,x,T,T, x,x,x,T, T,T,x,T, x,x,x,x, x,T,x,x, T,x,x,x, x,x,x,T, x,x,x,x, T,x,T,T, T,x,x,x, x,x,x,x, x,x,x,T, x,T,T,x, x,T,x,x, T,x,T,x, T,T,T,T, x,T,x,x, T,x,x,x, x,x,x,x, x,x,x,x, T,x,T,x, x,x,x,T, T,x,x,x, x,x,x,x, T,x,T,T, x,T,x,x, T,x,x,x, x,x,x,x, x,x,x,x, x,x,x},
		{x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, T,x,x,x, x,T,x,x, T,x,x,T, x,T,x,x, T,x,x,T, x,T,x,T, T,x,x,x, x,T,x,x, x,x,x},
		{x,x,x,x, x,x,x,T, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,T,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, T,T,T,T, x,x,x,x, x,x,x},
		{x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, T,T,T,T, x,x,x,x, x,x,x},
		{x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,x,T, x,x,x,x, x,x,T,x, T,x,x,x, x,x,x,x, x,x,x,x, x,x,x,x, x,x,T,x, x,T,x,x, x,x,x,x, x,x,x,x, x,x,x,x, T,x,x}

	};
} // end Parser


public class Errors {
	public int count = 0;                                    // number of errors detected
  public string errMsgFormat = "-- line {0} col {1}: {2}"; // 0=line, 1=column, 2=text
  public Queue encounteredErrors = new Queue();
	
	public void SynErr (int line, int col, int n) {
		string s;
		StringBuilder sb = new StringBuilder();
		switch (n) {
			case 0: s = "EOF expected"; break;
			case 1: s = "ident expected"; break;
			case 2: s = "intCon expected"; break;
			case 3: s = "realCon expected"; break;
			case 4: s = "charCon expected"; break;
			case 5: s = "stringCon expected"; break;
			case 6: s = "abstract expected"; break;
			case 7: s = "as expected"; break;
			case 8: s = "base expected"; break;
			case 9: s = "bool expected"; break;
			case 10: s = "break expected"; break;
			case 11: s = "byte expected"; break;
			case 12: s = "case expected"; break;
			case 13: s = "catch expected"; break;
			case 14: s = "char expected"; break;
			case 15: s = "checked expected"; break;
			case 16: s = "class expected"; break;
			case 17: s = "const expected"; break;
			case 18: s = "continue expected"; break;
			case 19: s = "decimal expected"; break;
			case 20: s = "default expected"; break;
			case 21: s = "delegate expected"; break;
			case 22: s = "do expected"; break;
			case 23: s = "double expected"; break;
			case 24: s = "else expected"; break;
			case 25: s = "enum expected"; break;
			case 26: s = "event expected"; break;
			case 27: s = "explicit expected"; break;
			case 28: s = "extern expected"; break;
			case 29: s = "false expected"; break;
			case 30: s = "finally expected"; break;
			case 31: s = "fixed expected"; break;
			case 32: s = "float expected"; break;
			case 33: s = "for expected"; break;
			case 34: s = "foreach expected"; break;
			case 35: s = "goto expected"; break;
			case 36: s = "if expected"; break;
			case 37: s = "implicit expected"; break;
			case 38: s = "in expected"; break;
			case 39: s = "int expected"; break;
			case 40: s = "interface expected"; break;
			case 41: s = "internal expected"; break;
			case 42: s = "is expected"; break;
			case 43: s = "lock expected"; break;
			case 44: s = "long expected"; break;
			case 45: s = "namespace expected"; break;
			case 46: s = "new expected"; break;
			case 47: s = "null expected"; break;
			case 48: s = "object expected"; break;
			case 49: s = "operator expected"; break;
			case 50: s = "out expected"; break;
			case 51: s = "override expected"; break;
			case 52: s = "params expected"; break;
			case 53: s = "private expected"; break;
			case 54: s = "protected expected"; break;
			case 55: s = "public expected"; break;
			case 56: s = "readonly expected"; break;
			case 57: s = "ref expected"; break;
			case 58: s = "return expected"; break;
			case 59: s = "sbyte expected"; break;
			case 60: s = "sealed expected"; break;
			case 61: s = "short expected"; break;
			case 62: s = "sizeof expected"; break;
			case 63: s = "stackalloc expected"; break;
			case 64: s = "static expected"; break;
			case 65: s = "string expected"; break;
			case 66: s = "struct expected"; break;
			case 67: s = "switch expected"; break;
			case 68: s = "this expected"; break;
			case 69: s = "throw expected"; break;
			case 70: s = "true expected"; break;
			case 71: s = "try expected"; break;
			case 72: s = "typeof expected"; break;
			case 73: s = "uint expected"; break;
			case 74: s = "ulong expected"; break;
			case 75: s = "unchecked expected"; break;
			case 76: s = "unsafe expected"; break;
			case 77: s = "ushort expected"; break;
			case 78: s = "usingKW expected"; break;
			case 79: s = "virtual expected"; break;
			case 80: s = "void expected"; break;
			case 81: s = "volatile expected"; break;
			case 82: s = "while expected"; break;
			case 83: s = "dot expected"; break;
			case 84: s = "colon expected"; break;
			case 85: s = "dblcolon expected"; break;
			case 86: s = "comma expected"; break;
			case 87: s = "scolon expected"; break;
			case 88: s = "lbrace expected"; break;
			case 89: s = "rbrace expected"; break;
			case 90: s = "lbrack expected"; break;
			case 91: s = "rbrack expected"; break;
			case 92: s = "lpar expected"; break;
			case 93: s = "rpar expected"; break;
			case 94: s = "tilde expected"; break;
			case 95: s = "condition expected"; break;
			case 96: s = "assgn expected"; break;
			case 97: s = "equal expected"; break;
			case 98: s = "nequal expected"; break;
			case 99: s = "not expected"; break;
			case 100: s = "and expected"; break;
			case 101: s = "andassgn expected"; break;
			case 102: s = "andand expected"; break;
			case 103: s = "or expected"; break;
			case 104: s = "orassgn expected"; break;
			case 105: s = "oror expected"; break;
			case 106: s = "xor expected"; break;
			case 107: s = "xorassgn expected"; break;
			case 108: s = "plus expected"; break;
			case 109: s = "plusassgn expected"; break;
			case 110: s = "inc expected"; break;
			case 111: s = "minus expected"; break;
			case 112: s = "minusassgn expected"; break;
			case 113: s = "dec expected"; break;
			case 114: s = "div expected"; break;
			case 115: s = "divassgn expected"; break;
			case 116: s = "times expected"; break;
			case 117: s = "timesassgn expected"; break;
			case 118: s = "mod expected"; break;
			case 119: s = "modassgn expected"; break;
			case 120: s = "gt expected"; break;
			case 121: s = "gteq expected"; break;
			case 122: s = "lt expected"; break;
			case 123: s = "lteq expected"; break;
			case 124: s = "ltlt expected"; break;
			case 125: s = "lshassgn expected"; break;
			case 126: s = "\"partial\" expected"; break;
			case 127: s = "\"yield\" expected"; break;
			case 128: s = "\"->\" expected"; break;
			case 129: s = "??? expected"; break;
			case 130: s = "invalid NamespaceMemberDeclaration"; break;
			case 131: s = "invalid Attributes"; break;
			case 132: s = "invalid TypeDeclaration"; break;
			case 133: s = "invalid TypeDeclaration"; break;
			case 134: s = "invalid TypeParameterConstraintsClause"; break;
			case 135: s = "invalid InterfaceMemberDeclaration"; break;
			case 136: s = "invalid InterfaceMemberDeclaration"; break;
			case 137: s = "invalid InterfaceMemberDeclaration"; break;
			case 138: s = "invalid IntegralType"; break;
			case 139: s = "invalid Type"; break;
			case 140: s = "invalid FormalParameterList"; break;
			case 141: s = "invalid ClassType"; break;
			case 142: s = "invalid ClassMemberDeclaration"; break;
			case 143: s = "invalid ClassMemberDeclaration"; break;
			case 144: s = "invalid StructMemberDeclaration"; break;
			case 145: s = "invalid StructMemberDeclaration"; break;
			case 146: s = "invalid StructMemberDeclaration"; break;
			case 147: s = "invalid StructMemberDeclaration"; break;
			case 148: s = "invalid StructMemberDeclaration"; break;
			case 149: s = "invalid StructMemberDeclaration"; break;
			case 150: s = "invalid StructMemberDeclaration"; break;
			case 151: s = "invalid StructMemberDeclaration"; break;
			case 152: s = "invalid StructMemberDeclaration"; break;
			case 153: s = "invalid StructMemberDeclaration"; break;
			case 154: s = "invalid Expression"; break;
			case 155: s = "invalid EventAccessorDeclarations"; break;
			case 156: s = "invalid EventAccessorDeclarations"; break;
			case 157: s = "invalid OverloadableOp"; break;
			case 158: s = "invalid AccessorDeclarations"; break;
			case 159: s = "invalid AccessorDeclarations"; break;
			case 160: s = "invalid AccessorDeclarations"; break;
			case 161: s = "invalid AccessorDeclarations"; break;
			case 162: s = "invalid InterfaceAccessors"; break;
			case 163: s = "invalid InterfaceAccessors"; break;
			case 164: s = "invalid NonGenericType"; break;
			case 165: s = "invalid LocalVariableDeclarator"; break;
			case 166: s = "invalid VariableInitializer"; break;
			case 167: s = "invalid Keyword"; break;
			case 168: s = "invalid AttributeArguments"; break;
			case 169: s = "invalid PrimitiveType"; break;
			case 170: s = "invalid PointerOrArray"; break;
			case 171: s = "invalid NonGenericClassType"; break;
			case 172: s = "invalid InternalClassType"; break;
			case 173: s = "invalid Statement"; break;
			case 174: s = "invalid EmbeddedStatement"; break;
			case 175: s = "invalid EmbeddedStatement"; break;
			case 176: s = "invalid EmbeddedStatement"; break;
			case 177: s = "invalid EmbeddedStatement"; break;
			case 178: s = "invalid StatementExpression"; break;
			case 179: s = "invalid ForInitializer"; break;
			case 180: s = "invalid CatchClauses"; break;
			case 181: s = "invalid ResourceAcquisition"; break;
			case 182: s = "invalid Unary"; break;
			case 183: s = "invalid AssignmentOperator"; break;
			case 184: s = "invalid SwitchLabel"; break;
			case 185: s = "invalid RelExpr"; break;
			case 186: s = "invalid RelExpr"; break;
			case 187: s = "invalid ShiftExpr"; break;
			case 188: s = "invalid Primary"; break;
			case 189: s = "invalid Primary"; break;
			case 190: s = "invalid Primary"; break;
			case 191: s = "invalid Literal"; break;

			default: s = "error " + n; break;
		}
		sb.AppendFormat(errMsgFormat, line, col, s);
		encounteredErrors.Enqueue(sb.ToString());
		count++;
	}

	public void SemErr (int line, int col, int n) {
		StringBuilder sb = new StringBuilder();
		sb.AppendFormat(errMsgFormat, line, col, ("error " + n));
		encounteredErrors.Enqueue(sb.ToString());
		count++;
	}

	public void Error (int line, int col, string s) {
		StringBuilder sb = new StringBuilder();
		sb.AppendFormat(errMsgFormat, line, col, s);
		encounteredErrors.Enqueue(sb.ToString());
		count++;
	}

	public void Exception (string s) {
		encounteredErrors.Enqueue(s);
		throw(new Exception("Fatal error while parsing!"));
	}
} // Errors

}