using System;
using System.Collections;
using System.Text;

namespace at.jku.ssw.Analyse {
	/// <summary>
	/// Diese Klasse ist zustndig fr die Speicherung der Anzahl der Schlsselwrter,
	/// Schleifenanzahl, etc.
	/// </summary>
	public class StatementCountInfo {
		//Anzahl der Schlsselwrter
		public Hashtable htKeywords;

		//Anzahl der cases in switch-Statements
		public int iCaseInSwitch;

		//Anzahl der Operatoren
		public Hashtable htOperators;

		//Anzahl der Statements
		public int iStatements;

		public StatementCountInfo() {
			htKeywords = new Hashtable();
			htOperators = new Hashtable();
		}

		/// <summary>
		/// Der + Operator wird am Ende eines Analyse-Durchganges bentigt, um aufsummierte
		/// StatementCountInfos fr Klassen, Namespaces und die ganze Datei zu berechnen.
		/// </summary>
		public static StatementCountInfo operator+ (StatementCountInfo sci1, StatementCountInfo sci2) {
			StatementCountInfo ret = new StatementCountInfo();
			//Addieren der Anzahlen der Schlsselwrter
			foreach (DictionaryEntry de in sci1.htKeywords) {
				ret.htKeywords[de.Key] = de.Value;
			}
			foreach (DictionaryEntry de in sci2.htKeywords) {
				if (ret.htKeywords.ContainsKey(de.Key)) {
					ret.htKeywords[de.Key] = (int) ret.htKeywords[de.Key] + (int) sci2.htKeywords[de.Key];
				}
				else ret.htKeywords[de.Key] = de.Value;
			}

			ret.iCaseInSwitch = sci1.iCaseInSwitch + sci2.iCaseInSwitch;
			//Addiren der Anzahlen der Operatoren
			foreach (DictionaryEntry de in sci1.htOperators) {
				ret.htOperators[de.Key] = de.Value;
			}
			foreach (DictionaryEntry de in sci2.htOperators) {
				if (ret.htOperators.ContainsKey(de.Key)) {
					ret.htOperators[de.Key] = (int) ret.htOperators[de.Key] + (int) sci2.htOperators[de.Key];
				}
				else ret.htOperators[de.Key] = de.Value;
			}
			//Addieren der Anzahl der Statements
			ret.iStatements = sci1.iStatements + sci2.iStatements;
			return ret;
		}

		/// <returns>Anzahl der Verzweigungen.</returns>
		public int getCntBranches() {
			int ret = 0;
			if (htKeywords.ContainsKey("if"))
				ret = (int) htKeywords["if"];
			if (htKeywords.ContainsKey("switch"))
				ret += (int) htKeywords["switch"];
			return ret;
		}

		/// <returns>Anzahl der for-Schleifen.</returns>
		public int getCntFor() {
			if (htKeywords.ContainsKey("for"))
				return (int) htKeywords["for"];
			else return 0;
		}

		/// <returns>Anzahl der foreach-Schleifen.</returns>
		public int getCntForeach() {
			if (htKeywords.ContainsKey("foreach"))
				return (int) htKeywords["foreach"];
			else return 0;
		}

		/// <returns>Anzahl der while-Schleifen.</returns>
		public int getCntWhile() {
			// da die do-while Schleife aus do und while besteht, 
			// muss man die do-while Schleifen von der Anzahl der
			// while-Schlsselwrter abziehen, um die Anzahl der 
			// while Schleifen zu erhalten.
			if (htKeywords.ContainsKey("while"))
				return (int) htKeywords["while"] - getCntDo();
			else return 0;
		}

		/// <returns>Anzahl der do-Schleifen.</returns>
		public int getCntDo() {
			if (htKeywords.ContainsKey("do"))
				return (int) htKeywords["do"];
			else return 0;
		}

		/// <returns>Anzahl der Schlsselwrter.</returns>
		public int getCntKeywords() {
			int ret = 0;
			foreach (DictionaryEntry de in htKeywords) ret += (int) de.Value;
			return ret;
		}

		/// <returns>Anzahl der Operatoren.</returns>
		public int getCntOperators() {
			int ret = 0;
			foreach (DictionaryEntry de in htOperators) ret += (int) de.Value;
			return ret;
		}

		public String ToString(StringBuilder tab) {
			StringBuilder sb = new StringBuilder();
			sb.Append(tab).AppendFormat("No. Statements:       {0}\n", iStatements);
			sb.Append(tab).AppendFormat("No. Branches:         {0}\n", getCntBranches());
			sb.Append(tab).AppendFormat("No. For-Loops:        {0}\n", getCntFor());
			sb.Append(tab).AppendFormat("No. Foreach-Loops:    {0}\n", getCntForeach());
			sb.Append(tab).AppendFormat("No. While-Loops:      {0}\n", getCntWhile());
			sb.Append(tab).AppendFormat("No. Do-While-Loops:   {0}\n", getCntDo());
			sb.Append(tab).AppendFormat("Keywords: {0}\n", getCntKeywords());
			foreach (DictionaryEntry de in htKeywords) {
				sb.Append(tab).Append(" ").Append(de.Value).Append(".   \t:\t").Append(de.Key).Append("\n");
			}
			sb.Append(tab).AppendFormat("Operators: {0}\n", getCntOperators());
			foreach (DictionaryEntry de in htOperators) {
				if ((int)de.Value > 0)
					sb.Append(tab).Append(" ").Append(de.Value).Append(".   \t:\t").Append(de.Key).Append("\n");
			}
			return sb.ToString();
		}
	}
}
